# -*- coding: utf-8 -*-
"""Breadcrumbs utility."""

from webhelpers2.html import literal

from .i18n import _


DEFAULT_ROOT_CHUNKS = 20


# =============================================================================
class Breadcrumbs(object):
    """User breadcrumb trail, current title page and back URL management.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str sep: (optional)
        Separator for brundcrumb trail.

    This class uses session and stores its history in
    ``session['breadcrumbs']``. It is a list of crumbs. Each crumb is a tuple
    such as ``(title, route_name, route_params, chunks_to_compare)``.
    """

    # -------------------------------------------------------------------------
    def __init__(self, request, sep=' ‣ '):
        """Constructor method."""
        self._request = request
        self._sep = sep

    # -------------------------------------------------------------------------
    def __call__(self, title, length=10, root_chunks=DEFAULT_ROOT_CHUNKS,
                 replace=None, anchor=None, forced_route=None,
                 compare_params=False):
        """Add a crumb in breadcrumb trail.

        :param str title:
            Current page title.
        :param int length: (default=10)
            Maximum crumb number. If 0, it keeps the current length.
        :param int root_chunks: (default=20)
            Number of route path chunks to compare to highlight a menu entry.
        :param str replace: (optional):
            If current path is ``replace``, this method call :meth:`pop` before
            any action.
        :param str anchor: (optional)
            Anchor to add.
        :param tuple forced_route: (optional)
            A tuple such as ``(route_name, route_params)`` to force the route.
        :param bool compare_params: (default=False)
            If ``True`` use route parameters to differentiate the current route
            from those currently in breadcrumbs.
        """
        # pylint: disable = too-many-arguments
        # Environment
        session = self._request.session
        if 'breadcrumbs' not in session:
            session['breadcrumbs'] = []
        if not length:
            length = len(session['breadcrumbs'])

        # Replace
        if replace and self.current_path() == replace:
            self.pop()

        # Scan old breadcrumb trail to find the right position
        route_name = (forced_route and forced_route[0]) or \
            (self._request.matched_route and self._request.matched_route.name)
        if route_name is None:
            session['breadcrumbs'].append((title, None, {}, root_chunks))
            return
        params = self._request.matchdict if forced_route is None \
            else forced_route[1]
        crumbs = []
        for crumb in session['breadcrumbs']:
            if len(crumbs) >= length - 1 \
               or (crumb[1] == route_name and not compare_params) \
               or (crumb[1] == route_name and crumb[2] == params):
                break
            crumbs.append(crumb)

        # Add new breadcrumb
        if anchor is not None:
            params['_anchor'] = anchor
        crumbs.append((title, route_name, params, root_chunks))
        session['breadcrumbs'] = crumbs

    # -------------------------------------------------------------------------
    def pop(self):
        """Pop last breadcrumb."""
        session = self._request.session
        if 'breadcrumbs' in session and len(session['breadcrumbs']) > 1:
            session['breadcrumbs'] = session['breadcrumbs'][0:-1]

    # -------------------------------------------------------------------------
    def trail(self):
        """Output XHTML breadcrumb trail.

        :rtype: str
        """
        if 'breadcrumbs' not in self._request.session \
           or len(self._request.session['breadcrumbs']) < 2:
            return literal('&nbsp;')

        translate = self._request.localizer.translate
        crumbs = []
        for crumb in self._request.session['breadcrumbs'][0:-1]:
            if crumb[1] is not None:
                crumbs.append('<a href="{path}">{label}</a>'.format(
                    path=self._request.route_path(crumb[1], **crumb[2]),
                    label=translate(crumb[0])))
            else:
                crumbs.append(translate(crumb[0]))
        crumbs.append(translate(self._request.session['breadcrumbs'][-1][0]))
        return literal(self._sep.join(crumbs))

    # -------------------------------------------------------------------------
    def crumb_trail(self):
        """Return a trail of crumbs to compare with menu entries. Each crumb
        is tuple of a list of path chunks and a comparison length.

        :rtype: list
        :return:
            A list of crumbs. Each crumb is a tuple such as
            ``([chunk1, chunk2,...], root_chunks)`` where ``root_chunks`` is a
            number of route path chunks to compare to highlight a menu entry
        """
        if self._request.matched_route is None:
            return []
        current = (
            self._request.current_route_path().partition('?')[0].split('/')[
                1:DEFAULT_ROOT_CHUNKS + 1], DEFAULT_ROOT_CHUNKS)
        if 'breadcrumbs' not in self._request.session:
            return [current]

        crumbs = []
        for crumb in self._request.session['breadcrumbs']:
            if crumb[1] is not None:
                path = self._request.route_path(
                    crumb[1], **crumb[2]).split('/')
                crumbs.append((path[1:crumb[3] + 1], crumb[3]))
        if not crumbs or crumbs[-1][0] != current[0]:
            crumbs.append(current)
        return crumbs

    # -------------------------------------------------------------------------
    def current_title(self):
        """Title of current page.

        :rtype: str
        """
        if 'breadcrumbs' not in self._request.session \
           or not self._request.session['breadcrumbs'] \
           or not self._request.session['breadcrumbs'][-1][1]:
            return self._request.localizer.translate(_('Home'))
        return self._request.localizer.translate(
            self._request.session['breadcrumbs'][-1][0])

    # -------------------------------------------------------------------------
    def current_route_name(self):
        """Route name of current page.

        :rtype: str
        """
        if 'breadcrumbs' not in self._request.session \
           or not self._request.session['breadcrumbs'] \
           or not self._request.session['breadcrumbs'][-1][1]:
            return 'home'
        return self._request.session['breadcrumbs'][-1][1]

    # -------------------------------------------------------------------------
    def current_path(self):
        """Path of current page.

        :rtype: str
        """
        if 'breadcrumbs' not in self._request.session \
           or not self._request.session['breadcrumbs'] \
           or not self._request.session['breadcrumbs'][-1][1]:
            return self._request.route_path('home')
        return self._request.route_path(
            self._request.session['breadcrumbs'][-1][1],
            **self._request.session['breadcrumbs'][-1][2])

    # -------------------------------------------------------------------------
    def back_title(self):
        """Output title of previous page.

        :rtype: str
        """
        if 'breadcrumbs' not in self._request.session \
           or len(self._request.session['breadcrumbs']) < 2 \
           or not self._request.session['breadcrumbs'][-2][1]:
            return self._request.localizer.translate(_('Home'))
        return self._request.localizer.translate(
            self._request.session['breadcrumbs'][-2][0])

    # -------------------------------------------------------------------------
    def back_path(self):
        """Output the path of previous page.

        :rtype: str
        """
        if 'breadcrumbs' not in self._request.session \
           or len(self._request.session['breadcrumbs']) < 2 \
           or not self._request.session['breadcrumbs'][-2][1]:
            return self._request.route_path('home')
        return self._request.route_path(
            self._request.session['breadcrumbs'][-2][1],
            **self._request.session['breadcrumbs'][-2][2])



# pylint: disable = import-outside-toplevel
"""Tests of ``views.mode`` functions."""

from . import DBUnitTestCase

MODE_TEST = ('test', (
    '{theme}/images/menu_test1.png', 'Test', None,
    ('user_view', {'user_id': 3}), None))


# =============================================================================
class UViewsModeModeView(DBUnitTestCase):
    """Unit test class for testing :func:`views.mode.mode_view`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:views.home.home_view]"""
        from collections import namedtuple
        from pyramid.httpexceptions import HTTPFound, HTTPNotFound
        from pyramid.httpexceptions import HTTPForbidden
        from ..modes import MODE_HOME, MODE_ADMIN
        from ..views.mode import mode_view

        self.configurator.add_route('home', '/')
        self.configurator.add_route('mode', '/mode/{mode_id}')
        self.configurator.add_route('settings_view', '/settings/view')
        self.configurator.add_route('user_view', '/user/view/{user_id}')
        self.request.matched_route = namedtuple('Route', 'name')(name='home')

        self.assertRaises(HTTPNotFound, mode_view, self.request)

        self.request.registry['modes'] = [MODE_HOME, MODE_TEST, MODE_ADMIN]
        self.request.session['modes'] = ['home', 'Home', None]
        self.request.matchdict = {'mode_id': 'foo'}
        self.assertRaises(HTTPNotFound, mode_view, self.request)

        self.request.matchdict = {'mode_id': 'test'}
        response = mode_view(self.request)
        self.assertIsInstance(response, HTTPFound)

        self.request.matchdict = {'mode_id': 'admin'}
        response = mode_view(self.request)
        self.assertIsInstance(response, HTTPFound)

        self.request.has_permission = lambda x: False
        self.assertRaises(HTTPForbidden, mode_view, self.request)



"""Route definitions."""

from pyramid.security import NO_PERMISSION_REQUIRED


# =============================================================================
def includeme(configurator):
    """Function to include routes.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    # Home
    configurator.add_route('home', '/')
    configurator.add_route('favicon', '/favicon.ico')
    configurator.add_route('robots', '/robots.txt')
    configurator.add_view(
        '.views.home.home_view', route_name='home',
        renderer=configurator.get_settings().get('site.home') or
        'chrysalio:Templates/home.pt')

    # Login/logout
    configurator.add_route('login', '/login')
    configurator.add_route('logout', '/logout')
    configurator.add_view(
        '.views.login.login', route_name='login',
        renderer=configurator.get_settings().get('site.login') or
        'chrysalio:Templates/login.pt', permission=NO_PERMISSION_REQUIRED)

    # Modes
    configurator.add_route('mode', '/mode/{mode_id}')

    # Panel
    configurator.add_route('panel_open', '/panel/open/{panel_id}')
    configurator.add_route('panel_close', '/panel/close/{panel_id}')

    # Attachment
    configurator.add_route('attachment', '/attachment/*path')

    # General settings
    configurator.add_route('settings_view', '/settings/view')
    configurator.add_route('settings_edit', '/settings/edit')

    # Backup
    configurator.add_route('backup', '/backup')

    # Profile
    configurator.add_route('profile_index', '/profile/index')
    configurator.add_route('profile_index_filter', '/profile/index/filter')
    configurator.add_route('profile_create', '/profile/create')
    configurator.add_route('profile_edit', '/profile/edit/{profile_id}')
    configurator.add_route('profile_view', '/profile/view/{profile_id}')

    # User
    configurator.add_route('user_index', '/user/index')
    configurator.add_route('user_index_filter', '/user/index/filter')
    configurator.add_route('user_create', '/user/create')
    configurator.add_route('user_edit', '/user/edit/{user_id}')
    configurator.add_route('user_view', '/user/view/{user_id}')
    configurator.add_route('user_account', '/user/account')
    configurator.add_route('user_password_forgot', '/user/password/forgot')
    configurator.add_route(
        'user_password_reset', '/user/password/reset/{user_id}/{token}')

    # Group
    configurator.add_route('group_index', '/group/index')
    configurator.add_route('group_index_filter', '/group/index/filter')
    configurator.add_route('group_create', '/group/create')
    configurator.add_route('group_edit', '/group/edit/{group_id}')
    configurator.add_route('group_view', '/group/view/{group_id}')



# -*- coding: utf-8 -*-
"""Class to divide large lists of items into pages."""

from string import Template
import re

from webhelpers2.html import literal, tags
from sqlalchemy.orm.query import Query

from ..lib.i18n import _
from ..includes.themes import theme_static_prefix

PAGE_DEFAULT_SIZE = 80
PAGE_SIZES = (('', ' '), 10, 20, 40, 80, 160, 320, 640, ('0', '∞'))


# =============================================================================
def sortable_column(
        request, label, sort, current_sorting=None):
    """Output a header of column with `sort up` and `sort down` buttons.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str label:
        Label of column.
    :param str sort:
        Sort criteria.
    :param str current_sorting: (optional)
        Default current sorting.
    :rtype: webhelpers2.html.literal
    """
    current = request.params.get('sort') or current_sorting
    query_string = {}
    if request.GET:
        query_string.update(request.GET)

    html = '<a title="{0}"'.format(request.localizer.translate(
        _('Sort by ${l}', {'l': label.lower()})))
    if current and sort == current[1:]:
        html += current[0] == '+' and \
                ' class="cioSortAsc"' or ' class="cioSortDesc"'
    if current and sort == current[1:] and current[0] == '+':
        query_string['sort'] = '-{0}'.format(sort)
    else:
        query_string['sort'] = '+{0}'.format(sort)
    html += ' href="{0}"'.format(
        request.current_route_path(_query=query_string))
    html += '>{0}</a>'.format(label)

    return literal(html)


# =============================================================================
class Paging(list):
    """Divide large lists of items into pages.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str paging_id:
        Paging ID.
    :type  collection: list-like object
    :param collection:
        Collection object being paged through.
    :param dict params: (optional)
        A dictionary with the following keys: ``page_size``, ``page``,
        ``sort`` and ``display``.
    :param int item_count: (optional)
        Number of items in the collection.
    :param tuple page_sizes: (optional)
        List of suggested page sizes.

    This class uses the following parameters in request: ``page_size``,
    ``page``, ``sort`` and ``display``.

    It stores its information and filters definitions in ``session['paging']``.
    This structure looks like: ``session['paging'] = (page_default_size,
    {'page_id1': {'page_size': 80, 'page': 3, 'sort': 'name',
    'display': 'cards'}, 'page_id2': {...}, ...})``
    """
    # pylint: disable = too-many-instance-attributes

    # -------------------------------------------------------------------------
    def __init__(self, request, paging_id, collection, params=None,
                 item_count=None):
        """Constructor method."""
        # Update paging session
        if params is None:
            params = self.params(request, paging_id)

        # Initialize variables
        self._request = request
        self.paging_id = paging_id
        full_list = collection
        if isinstance(collection, Query):
            full_list = _SQLAlchemyQuery(collection)
        self.item_count = item_count if item_count is not None else \
            len(full_list)
        self.page_count = ((self.item_count - 1) // (params['page_size'])) + 1\
            if params['page_size'] else 1
        if params['page'] > self.page_count:
            params['page'] = self.page_count
        self.page = max(1, params['page'])
        self.items = []
        self.page_size = params['page_size']
        self.display = params['display']

        # Compute the item list
        if self.item_count > 0:
            if self.page_size:
                first_item = (self.page - 1) * self.page_size + 1
                last_item = min(
                    first_item + self.page_size - 1, self.item_count)
            else:
                first_item = 1
                last_item = self.item_count
            try:
                self.items = full_list[first_item - 1:last_item]
            except TypeError:
                raise TypeError("You can't use type %s!" % type(full_list))

        list.__init__(self, self.items)

    # -------------------------------------------------------------------------
    @classmethod
    def params(cls, request, paging_id, default_sorting=None,
               default_display='cards'):
        """Return current paging parameters: page number, page size and
        sorting.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str paging_id:
            Paging ID.
        :param str default_sorting: (optional)
            Default sorting.
        :param str default_display: ('cards' or 'list', default='cards')
            Default display.
        :rtype: dict
        :return:
            The paging dictionary. See :class:`~.paging.Paging` class.
        """
        if request is None:
            return {
                'page_size': PAGE_DEFAULT_SIZE, 'page': 1,
                'sort': default_sorting, 'display': default_display}

        if 'paging' not in request.session:
            request.session['paging'] = (
                request.registry['settings']['page-size']
                if 'settings' in request.registry else PAGE_DEFAULT_SIZE, {})
        if paging_id not in request.session['paging'][1]:
            request.session['paging'][1][paging_id] = {
                'page_size': request.session['paging'][0],
                'page': 1, 'sort': default_sorting, 'display': default_display}
        params = request.session['paging'][1][paging_id]

        if 'page_size' in request.params \
           and request.params['page_size'].strip():
            params['page_size'] = int(request.params['page_size'])
        if 'page' in request.params and request.params['page'].isdigit():
            params['page'] = max(1, int(request.params['page']))
        if request.params.get('sort'):
            params['sort'] = request.params['sort']
        if params['sort'] is None:
            params['sort'] = default_sorting
        if request.params.get('display'):
            params['display'] = request.params['display']

        if request.POST:
            request.session['paging'][1][paging_id] = {
                'page_size': params['page_size'],
                'page': params['page'],
                'sort': params['sort'],
                'display': params['display']}
            params = request.session['paging'][1][paging_id]

        return params

    # -------------------------------------------------------------------------
    @classmethod
    def get_sort(cls, request, paging_id):
        """Retrieve the sort criteria from the session.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str paging_id:
            ID of the paging.
        :rtype: :class:`str` or ``None``
        """
        if 'paging' not in request.session or \
           paging_id not in request.session['paging'][1]:
            return None
        return request.session['paging'][1][paging_id]['sort']

    # -------------------------------------------------------------------------
    @classmethod
    def get_page(cls, request, paging_id):
        """Retrieve the current page number from the session.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str paging_id:
            ID of the paging.
        :rtype: int
        """
        if 'paging' not in request.session or \
           paging_id not in request.session['paging'][1]:
            return 1
        return request.session['paging'][1][paging_id]['page']

    # -------------------------------------------------------------------------
    def get_item(self, field_id, value):
        """Retrieve the first item whose field ``field_id`` has the value
        ``value``.

        :param str field_id:
            Name of the item field to search.
        :param value:
            Value to use to find the item.
        :rtype: :class:`dict` or ``None``
        """
        try:
            return next((k for k in self if k[field_id] == value))
        except (StopIteration, AttributeError, TypeError):
            return None

    # -------------------------------------------------------------------------
    def set_current_ids(self, field_id):
        """Save in ``session['paging'][1][self.paging_id]['current_ids']``
        the IDs of the items of the page.

        :param str field_id:
            Name of the item field used to store the IDs.
        """
        if self._request is None or \
           'paging' not in self._request.session or \
           self.paging_id not in self._request.session['paging'][1]:
            return
        self._request.session['paging'][1][self.paging_id][
            'current_ids'] = [getattr(k, field_id) for k in self]

    # -------------------------------------------------------------------------
    def pager(self, pager_format='~3~',
              symbol_first='&lt;&lt;', symbol_first_off='',
              symbol_last='&gt;&gt;', symbol_last_off='',
              symbol_previous='&lt;', symbol_previous_off='',
              symbol_next='&gt;', symbol_next_off=''):
        """Return string with links to other pages (e.g. '1 .. 5 6 7 [8] 9 10
        11 .. 50').

        :param str pager_format: (default='~3~')
            Format string that defines how the pager is rendered.
        :param str symbol_first: (default='&lt;&lt;')
            String to be displayed as the text for the $link_first link above.
        :param str symbol_last: (default='&gt;&gt;')
            String to be displayed as the text for the $link_last link above.
        :param str symbol_previous: (default='&lt;')
            String to be displayed as the text for the $link_previous link
            above.
        :param str symbol_next: (default='&gt;')
            String to be displayed as the text for the $link_next link above.

        Format string that defines how the pager is rendered. The string
        can contain the following $-tokens that are substituted by the
        string.Template module:

        - $first_page: number of first reachable page
        - $last_page: number of last reachable page
        - $page: number of currently selected page
        - $page_count: number of reachable pages
        - $items_per_page: maximal number of items per page
        - $first_item: index of first item on the current page
        - $last_item: index of last item on the current page
        - $item_count: total number of items
        - $link_first: link to first page (unless this is first page)
        - $link_first_off: link to first page (for first page)
        - $link_last_off: link to last page (for last page)
        - $link_previous: link to previous page (unless this is first page)
        - $link_previous_off: link to previous page (for first page)
        - $link_next_off: link to next page (for last page)

        To render a range of pages the token '~3~' can be used. The number sets
        the radius of pages around the current page.
        Example for a range with radius 3: '1 .. 5 6 7 [8] 9 10 11 .. 50'
        """
        # pylint: disable = too-many-arguments
        if self._request is None:
            return ''

        # Replace ~...~ in token format by range of pages
        result = re.sub(r'~(\d+)~', self._range, pager_format)

        # Interpolate '$' variables
        items_per_page = int(self._request.params['page_size']) \
            if self._request.params.get('page_size') else \
            self._request.session['paging'][1][self.paging_id]['page_size']
        query_string = self._request.GET.copy()
        first_item = min((self.page - 1) * self.page_size + 1, self.item_count)
        last_item = min(first_item + items_per_page - 1, self.item_count) \
            if self.page_size else self.item_count

        return literal(Template(result).safe_substitute({
            'first_page': 1,
            'last_page': self.page_count,
            'page': self.page,
            'page_count': self.page_count,
            'items_per_page': items_per_page,
            'first_item': first_item,
            'last_item': last_item,
            'item_count': self.item_count,
            'link_first': self.page > 1 and self._link(
                query_string, symbol_first, 1) or symbol_first_off,
            'link_last': self.page < self.page_count and self._link(
                query_string, symbol_last, self.page_count) or symbol_last_off,
            'link_previous': self.page > 1 and self._link(
                query_string, symbol_previous,
                self.page - 1) or symbol_previous_off,
            'link_next': self.page < self.page_count and self._link(
                query_string, symbol_next, self.page + 1) or symbol_next_off}))

    # -------------------------------------------------------------------------
    def pager_top(self, image='{theme}/images/paging_go_{action}.png'):
        """Output a string with links to first, previous, next and last pages.

        :param str image: (default={theme}/images/paging_go_{action}.png)
            Pattern for the route to navigation button images.
        :rtye: str
        """
        theme = theme_static_prefix(self._request)
        image = '<img src="{0}" alt="{1}"/>'.format(
            image.format(theme=theme, action='%s'), '%s')
        return self.pager(
            '$link_first $link_previous '
            '<span>$first_item &ndash; $last_item</span> / $item_count '
            '$link_next $link_last',
            symbol_first=literal(image % ('first', 'First')),
            symbol_first_off=literal(image % ('first_off', 'First')),
            symbol_previous=literal(image % ('previous', 'Previous')),
            symbol_previous_off=literal(image % ('previous_off', 'Previous')),
            symbol_next=literal(image % ('next', 'Next')),
            symbol_next_off=literal(image % ('next_off', 'Next')),
            symbol_last=literal(image % ('last', 'Last')),
            symbol_last_off=literal(image % ('last_off', 'Last'))) or \
            literal('&nbsp;')

    # -------------------------------------------------------------------------
    def pager_bottom(self):
        """Output a string with links to some previous and next pages.

        :rtype: str
        """
        return self.pager() or literal('&nbsp;')

    # -------------------------------------------------------------------------
    def display_modes(self, image='{theme}/images/paging_{display}.png'):
        """Output buttons to switch between cards and list mode.

        :param str image: (default={theme}/images/paging_{action}.png)
            Pattern for the route to navigation button images.
        :rtype: str
        """
        if self._request is None:
            return ''

        theme = theme_static_prefix(self._request)
        image = '<img src="{0}" alt="{1}" title="{2}"/>'.format(
            image.format(theme=theme, display='%s'), '%s', '%s')
        query_string = self._request.GET.copy()
        translate = self._request.localizer.translate
        html = literal('<span class="cioPagingDisplay">')

        if self.display == 'cards':
            html += literal(image % (
                'cards_off', 'Cards', translate(_('Display as Cards'))))
        else:
            query_string.update({'display': 'cards'})
            html += self._link(
                query_string, literal(image % (
                    'cards', 'Cards', translate(_('Display as Cards')))))
        html += ' '

        if self.display == 'list':
            html += literal(
                image % ('list_off', 'List', translate(_('Display as List'))))
        else:
            query_string.update({'display': 'list'})
            html += self._link(
                query_string, literal(
                    image % ('list', 'List', translate(_('Display as List')))))

        html += literal('</span>')
        return html

    # -------------------------------------------------------------------------
    @classmethod
    def navigator(cls, request, paging_id, item_id, url,
                  image='{theme}/images/paging_go_{action}.png'):
        """Return a piece of HTML to go to the previous and the next item.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str paging_id:
            Paging ID.
        :param str item_id:
            ID of the current item.
        :param str url:
            Pattern for the URL of previous and next button.
        :param str image: (default={theme}/images/paging_{action}.png)
            Pattern for the route to navigation button images.
        :rtype: str
        """
        if request is None or 'paging' not in request.session or \
           paging_id not in request.session['paging'][1] or \
           'current_ids' not in request.session['paging'][1][paging_id]:
            return ''

        current_ids = request.session['paging'][1][paging_id]['current_ids']
        if item_id not in current_ids:
            return ''
        index = current_ids.index(item_id)
        theme = theme_static_prefix(request)
        image = '<img src="{0}" alt="{1}"/>'.format(
            image.format(theme=theme, action='%s'), '%s')

        html = '<span class="cioPagingNavigator">'
        if index:
            html += '<a href="{0}">{1}</a>'.format(
                url.replace('_ID_', str(current_ids[index - 1])),
                image % ('previous', 'Previous'))
        else:
            html += image % ('previous_off', 'Previous')
        html += ' <span>{0}</span> / {1} '.format(index + 1, len(current_ids))
        if index < len(current_ids) - 1:
            html += '<a href="{0}">{1}</a>'.format(
                url.replace('_ID_', str(current_ids[index + 1])),
                image % ('next', 'Next'))
        else:
            html += image % ('next_off', 'Next')
        html += '</span>'

        return literal(html)

    # -------------------------------------------------------------------------
    def sortable_column(self, label, sort):
        """Output a header of column with `sort up` and `sort down` buttons.

        See :func:`sortable_column`.

        :param str label:
             Label of column.
        :param str sort:
             Sort criteria.
        :rtype: webhelpers2.html.literal
        """
        if self._request is None:
            return '&nbsp;'
        return sortable_column(
            self._request, label, sort,
            self._request.session['paging'][1][self.paging_id]['sort'])

    # -------------------------------------------------------------------------
    def _range(self, regex_match):
        """Return range of linked pages (e.g. '1 2 [3] 4 5 6 7 8').

        :type  regex_match: re.Match
        :param regex_match:
            A regular expressions match object containing the radius of linked
            pages around the current page in regex_match.group(1) as a string.
        :rtype: str
        """
        query_string = self._request.GET.copy()
        radius = int(regex_match.group(1))
        leftmost_page = max(1, self.page - radius)
        rightmost_page = min(self.page + radius, self.page_count)
        items = []

        if self.page != 1 and leftmost_page > 1:
            items.append(self._link(query_string, '1', 1))
        if leftmost_page > 2:
            items.append('..')
        for page in range(leftmost_page, rightmost_page + 1):
            if page == self.page:
                items.append('<span>%d</span>' % page)
            else:
                items.append(self._link(query_string, str(page), page))
        if rightmost_page < self.page_count - 1:
            items.append('..')
        if self.page != self.page_count and rightmost_page < self.page_count:
            items.append(self._link(
                query_string, str(self.page_count), self.page_count))

        return ' '.join(items)

    # -------------------------------------------------------------------------
    def _link(self, query_string, label, page_number=None):
        """Create an A-HREF tag.

        :param dict query_string:
            The current query string in a dictionary.
        :param str label:
            Text to be printed in the A-HREF tag.
        :param int page_number: (optional)
            Number of the page that the link points to.
        """
        if page_number:
            query_string.update({'page': page_number})
        return tags.link_to(
            label, self._request.current_route_path(_query=query_string))


# =============================================================================
class _SQLAlchemyQuery(object):
    """Iterable that allows to get slices from an SQLAlchemy Query object."""
    # pylint: disable = too-few-public-methods

    # -------------------------------------------------------------------------
    def __init__(self, query):
        """Contructor method."""
        self.query = query

    # -------------------------------------------------------------------------
    def __getitem__(self, records):
        """Implement evaluation of self[key]."""
        if not isinstance(records, slice):  # pragma: nocover
            raise Exception('__getitem__ without slicing not supported')
        return self.query[records]

    # -------------------------------------------------------------------------
    def __len__(self):
        """Implement the built-in function len()."""
        return self.query.count()



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``includes.themes`` classes and functions."""

from os.path import normpath, dirname, join
from unittest import TestCase

from pyramid.testing import setUp, tearDown
from pyramid.testing import DummyRequest


# =============================================================================
class ThemeTestCase(TestCase):
    """Base class for testing themes."""

    configurator = None
    layout = normpath(join(dirname(__file__), '..', 'Templates', 'layout.pt'))

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.configurator = setUp(settings={
            'theme.roots': 'chrysalio:Themes',
            'theme.patterns': 'Default, Alternative*'})
        self.configurator.registry['settings'] = {'theme': ''}

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        tearDown()


# =============================================================================
class UIncludesThemesIncludeme(ThemeTestCase):
    """Unit test class for :func:`includes.themes.includeme`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.includeme]"""
        self.configurator.include('..includes.themes')
        self.assertIn('themes', self.configurator.registry)


# =============================================================================
class UIncludesThemesLoadThemes(ThemeTestCase):
    """Unit test class for :func:`includes.themes.load_themes`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.load_themes]"""
        from ..includes.themes import load_themes

        load_themes(self.configurator)
        self.assertIn('themes', self.configurator.registry)
        themes = self.configurator.registry['themes']
        self.assertIn('Default', themes)
        self.assertIn('Alternative', themes)
        self.assertIn('path', themes['Default'])
        self.assertIn('name', themes['Default'])
        self.assertIn('layout', themes['Default'])
        self.assertIsInstance(themes['Default']['name'], dict)
        self.assertEqual(themes['Default']['name']['en'], 'Chrysalio')
        self.assertEqual(themes['Default']['name']['fr'], 'Chrysalio')
        self.assertEqual(
            themes['Alternative']['name']['en'], 'Chrysalio Alternative')


# =============================================================================
class UIncludesThemesLoadThemesMissingTemplates(TestCase):
    """Unit test class for :func:`includes.themes.load_themes` with missing
    [Templates] section."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.configurator = setUp(settings={
            'theme.roots': 'chrysalio:tests/Themes',
            'theme.patterns': 'Test1'})

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        tearDown()

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.load_themes] section [Templates] is missing"""
        from ..includes.themes import load_themes

        self.assertRaises(SystemExit, load_themes, self.configurator)


# =============================================================================
class UIncludesThemesLoadThemesMissingLayout(TestCase):
    """Unit test class for :func:`includes.themes.load_themes` with missing
    layout."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.configurator = setUp(settings={
            'theme.roots': 'chrysalio:tests/Themes',
            'theme.patterns': 'Test2'})

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        tearDown()

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.load_themes] layout is missing"""
        from ..includes.themes import load_themes

        self.assertRaises(SystemExit, load_themes, self.configurator)


# =============================================================================
class UIncludesThemesLoadThemesMissingCss(TestCase):
    """Unit test class for :func:`includes.themes.load_themes` with missing
    CSS."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.configurator = setUp(settings={
            'theme.roots': 'chrysalio:tests/Themes',
            'theme.patterns': 'Test3'})

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        tearDown()

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.load_themes] CSS is missing"""
        from ..includes.themes import load_themes

        self.assertRaises(SystemExit, load_themes, self.configurator)


# =============================================================================
class UIncludesThemesLoadThemesEmptyCss(TestCase):
    """Unit test class for :func:`includes.themes.load_themes` with empty
    CSS."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.configurator = setUp(settings={
            'theme.roots': 'chrysalio:tests/Themes',
            'theme.patterns': 'Test4'})

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        tearDown()

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.load_themes] CSS is empty"""
        from ..includes.themes import load_themes

        self.assertRaises(SystemExit, load_themes, self.configurator)


# =============================================================================
class UIncludesThemesLoadThemesUnknownCss(TestCase):
    """Unit test class for :func:`includes.themes.load_themes` with unknown
    CSS."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.configurator = setUp(settings={
            'theme.roots': 'chrysalio:tests/Themes',
            'theme.patterns': 'Test5'})

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        tearDown()

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.load_themes] CSS is unknown"""
        from ..includes.themes import load_themes

        self.assertRaises(SystemExit, load_themes, self.configurator)


# =============================================================================
class UIncludesThemesLoadThemesUnknownDefault(TestCase):
    """Unit test class for :func:`includes.themes.load_themes` with unknown
    default theme."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.configurator = setUp(settings={
            'theme.roots': 'chrysalio:Themes',
            'theme.patterns': 'Default, Alternative*'})
        self.configurator.registry['settings'] = {'theme': 'foo'}

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        tearDown()

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.load_themes] default theme is unknown"""
        from ..includes.themes import load_themes

        self.assertRaises(SystemExit, load_themes, self.configurator)


# =============================================================================
class UIncludesThemesCreateDefaultTheme(ThemeTestCase):
    """Unit test class for ``includes.themes.create_default_theme``."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.create_default_theme]"""
        from ..includes.themes import create_default_theme

        self.assertNotIn('themes', self.configurator.registry)
        create_default_theme(self.configurator, 'chrysalio')
        self.assertIn('themes', self.configurator.registry)
        themes = self.configurator.registry['themes']
        self.assertIn('path', themes[''])
        self.assertIn('name', themes[''])
        self.assertIn('layout', themes[''])
        self.assertEqual(themes['']['layout'], self.layout)


# =============================================================================
class UIncludesThemesThemeTemplate(ThemeTestCase):
    """Unit test class for ``includes.themes.theme_template``."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.theme_template]"""
        from ..includes.themes import theme_template

        self.configurator.include('..includes.themes')
        request = DummyRequest()
        path = theme_template(request, 'layout')
        self.assertEqual(path, self.layout)


# =============================================================================
class UIncludesThemesThemeStaticPrefix(ThemeTestCase):
    """Unit test class for ``includes.themes.theme_static_prefix``."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.theme_static_prefix]"""
        from ..includes.themes import theme_static_prefix

        self.configurator.include('..includes.themes')

        prefix = theme_static_prefix(None)
        self.assertEqual(prefix, '')

        request = DummyRequest()
        prefix = theme_static_prefix(request)
        self.assertEqual(prefix, '/theme/default')


# =============================================================================
class UIncludesThemesThemeHasStatic(ThemeTestCase):
    """Unit test class for ``includes.themes.theme_has_static``."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.themes.theme_has_static]"""
        from ..includes.themes import theme_has_static

        request = DummyRequest()
        self.configurator.include('..includes.themes')
        self.assertTrue(theme_has_static(request, 'css'))
        self.assertFalse(theme_has_static(request, 'css+'))



# -*- coding: utf-8 -*-
"""User group management view callables."""

from sqlalchemy import desc
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import FlushError

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound, HTTPNotFound

from ..lib.i18n import _
from ..lib.log import log_info
from ..lib.form import get_action, Form
from ..lib.filter import Filter
from ..lib.paging import PAGE_SIZES, Paging
from ..lib.utils import make_id
from ..lib.tabset import Tabset
from ..lib.attachment import attachment_url, attachment_update
from ..includes.themes import theme_static_prefix
from ..models.populate import xml2db, db2web, web2db
from ..models.dbprofile import DBProfile
from ..models.dbuser import DBUser
from ..models.dbgroup import DBGroup, DBGroupUser, DBGroupProfile
from . import BaseView


# =============================================================================
class GroupView(BaseView):
    """Class to manage user group views.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    _DBGroup = DBGroup
    _xml2db = (xml2db,)

    # -------------------------------------------------------------------------
    @view_config(
        route_name='group_index',
        renderer='chrysalio:Templates/group_index.pt',
        permission='group-view')
    @view_config(route_name='group_index', renderer='json', xhr=True)
    def index(self):
        """List all user groups."""
        # Ajax
        i_creator = self._request.has_permission('group-create')
        if self._request.is_xhr:
            if i_creator:
                self._import_groups()
            return {}

        # Action
        action, items = get_action(self._request)
        if action[:4] == 'del!' and i_creator:
            self._delete_groups(items)
        elif action == 'imp!' and i_creator:
            self._import_groups()
        elif action[:4] == 'exp!':
            action = self._groups2response(items)
            if action:
                return action

        # Filter
        paging_id = 'groups'
        pfilter = Filter(
            self._request, paging_id, (
                ('group_id', _('Identifier'), False, ''),
                ('i18n_label', _('Label'), False, '')),
            remove=action[:4] == 'crm!' and action[4:] or None)

        # Paging
        defaults = Paging.params(self._request, paging_id, '+group_id')
        dbquery = pfilter.sql(
            self._request.dbsession.query(self._DBGroup), 'groups')
        oby = getattr(self._DBGroup, defaults['sort'][1:])
        dbquery = dbquery.order_by(
            desc(oby) if defaults['sort'][0] == '-' else oby)
        group_paging = Paging(self._request, paging_id, dbquery, defaults)
        group_paging.set_current_ids('group_id')

        # Form & completed action
        form = Form(self._request, defaults=defaults)
        form.forget('filter_value')
        if action and action[3] == '!':
            action = ''

        # Breadcrumbs
        self._request.breadcrumbs(_('Groups'), 2)

        return {
            'action': action, 'items': items, 'form': form, 'pfilter': pfilter,
            'paging': group_paging, 'PAGE_SIZES': PAGE_SIZES,
            'download_max_size': self._request.registry['settings'][
                'download-max-size'],
            'i_creator': i_creator,
            'i_editor': self._request.has_permission('group-edit'),
            'has_attachments': bool(
                self._request.registry.settings.get('attachments')),
            'attachment_url': attachment_url}

    # -------------------------------------------------------------------------
    @view_config(route_name='group_index_filter', renderer='json', xhr=True)
    def index_filter(self):
        """Return a list to autocomplete a filter field."""
        print(Filter.sql_autocomplete(self._request, self._DBGroup))
        return Filter.sql_autocomplete(self._request, self._DBGroup)

    # -------------------------------------------------------------------------
    @view_config(
        route_name='group_view',
        renderer='chrysalio:Templates/group_view.pt',
        permission='group-view')
    def view(self):
        """Show user group."""
        dbgroup = self._get_group()
        picture = self._request.registry.settings.get('attachments') and (
            attachment_url(
                self._request, dbgroup.attachments_dir,
                dbgroup.attachments_key, dbgroup.picture) or
            '{0}/images/group_picture.svg'.format(
                theme_static_prefix(self._request)))

        action = get_action(self._request)[0]
        if action == 'exp!':
            action = self._groups2response((dbgroup.group_id,))
            if action:
                return action

        # User paging
        user_paging, defaults, user_filter = self._user_paging(
            action, dbgroup)

        # Form
        form = Form(self._request, defaults=defaults)
        form.forget('filter_value')

        # Breadcrumbs
        label = dbgroup.label(self._request)
        self._request.breadcrumbs(
            _('Group "${l}"', {'l': label}),
            replace=self._request.route_path(
                'group_edit', group_id=dbgroup.group_id))

        return {
            'form': form, 'label': label, 'user_filter': user_filter,
            'user_paging': user_paging, 'tabset': Tabset(
                self._request, 'tabGroup',
                dbgroup.settings_tabs(self._request)),
            'picture': picture, 'dbgroup': dbgroup,
            'navigator': Paging.navigator(
                self._request, 'groups', dbgroup.group_id,
                self._request.route_path('group_view', group_id='_ID_')),
            'i_editor': self._request.has_permission('group-edit')}

    # -------------------------------------------------------------------------
    @view_config(
        route_name='group_create',
        renderer='chrysalio:Templates/group_edit.pt',
        permission='group-create')
    @view_config(
        route_name='group_edit',
        renderer='chrysalio:Templates/group_edit.pt',
        permission='group-edit')
    @view_config(
        route_name='group_edit', renderer='json', xhr=True,
        permission='group-edit')
    def edit(self):
        """Create or edit a user group."""
        # Initialization
        dbgroup = self._get_group() \
            if 'group_id' in self._request.matchdict else None
        # Ajax
        if self._request.is_xhr:
            if dbgroup is not None:
                dbgroup.attachments_key, dbgroup.picture = \
                    attachment_update(
                        self._request, dbgroup.attachments_dir,
                        dbgroup.attachments_key,
                        self._request.POST['picture'],
                        replace=dbgroup.picture,
                        prefix=dbgroup.group_id[:12])
                log_info(
                    self._request, 'group_update_picture',
                    dbgroup.group_id)
            return {}

        # User paging
        action = get_action(self._request)[0]
        user_paging, defaults, user_filter = self._user_paging(action)

        # Profiles
        profiles = {
            k.profile_id: (
                k.label(self._request), k.description(self._request))
            for k in self._request.dbsession.query(DBProfile)}

        # Form and action
        form = Form(
            self._request,
            *self._DBGroup.settings_schema(
                self._request, defaults, profiles, dbgroup),
            obj=dbgroup)
        if action == 'pct!' and dbgroup is not None:
            dbgroup.attachments_key, dbgroup.picture = \
                attachment_update(
                    self._request, dbgroup.attachments_dir,
                    dbgroup.attachments_key, self._request.POST['picture'],
                    replace=dbgroup.picture,
                    prefix=dbgroup.group_id[:12])
            log_info(
                self._request, 'group_update_picture',
                dbgroup.group_id)
        elif action == 'sav!' and form.validate():
            dbgroup = self._save(dbgroup, profiles, form.values)
            if dbgroup is not None:
                dbuser = self._request.dbsession.query(DBUser).filter_by(
                    user_id=self._request.session['user']['user_id']).first()
                dbuser.set_session(self._request)
                if 'group_id' not in self._request.matchdict:
                    self._request.breadcrumbs.pop()
                log_info(
                    self._request,
                    'group_id' in self._request.matchdict and
                    'group_edit' or 'group_create', dbgroup.group_id)
                return HTTPFound(self._request.route_path(
                    'group_view', group_id=dbgroup.group_id))
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        # Picture
        picture = \
            dbgroup \
            and self._request.registry.settings.get('attachments') \
            and (attachment_url(
                self._request, DBGroup.attachments_dir,
                dbgroup.attachments_key, dbgroup.picture) or
                 '{0}/images/group_picture.svg'.format(
                     theme_static_prefix(self._request)))

        # Breadcrumbs
        label = dbgroup and dbgroup.label(self._request)
        if not dbgroup:
            self._request.breadcrumbs(_('Group Creation'))
        else:
            self._request.breadcrumbs(
                _('Group "${l}" Edition', {'l': label}),
                replace=self._request.route_path(
                    'group_view', group_id=dbgroup.group_id))
        return {
            'form': form, 'dbgroup': dbgroup or self._DBGroup,
            'action': action, 'label': label, 'profiles': profiles,
            'user_filter': user_filter, 'user_paging': user_paging,
            'picture': picture, 'tabset': Tabset(
                self._request, 'tabGroup',
                self._DBGroup.settings_tabs(self._request))}

    # -------------------------------------------------------------------------
    def _get_group(self):
        """Return the SqlAlchemy object of the selected user group or raise
        an HTTPNotFound exception.

        :rtype: .models.dbgroup.DBGroup
        """
        group_id = self._request.matchdict['group_id']
        dbgroup = self._request.dbsession.query(self._DBGroup).filter_by(
            group_id=group_id).first()
        if dbgroup is None:
            raise HTTPNotFound()
        return dbgroup

    # -------------------------------------------------------------------------
    def _delete_groups(self, group_ids):
        """Delete groups.

        :param list group_ids:
            List of group IDs to delete.
        """
        deleted = []
        for dbgroup in self._request.dbsession.query(self._DBGroup).filter(
                self._DBGroup.group_id.in_(group_ids)):
            deleted.append(dbgroup.group_id)
            self._request.dbsession.delete(dbgroup)
        if deleted:
            log_info(self._request, 'group_delete', ' '.join(deleted))

    # -------------------------------------------------------------------------
    def _import_groups(self):
        """Import groups."""
        # Get current IDs
        group_ids = {k[0] for k in self._request.dbsession.query(
            DBGroup.group_id)}

        # Update database
        web2db(self._request, self._xml2db[0], 'group')

        # Get new IDs
        group_ids = {k[0] for k in self._request.dbsession.query(
            DBGroup.group_id)} - group_ids
        if group_ids:
            log_info(self._request, 'group_import', ' '.join(group_ids))

    # -------------------------------------------------------------------------
    def _groups2response(self, group_ids):
        """Export user groups as an XML file embedded in a Pyramid response.

        :param list group_ids:
            List of group IDs to export.
        :rtype: :class:`pyramid.response.Response` or ``''``
        """
        dbitems = tuple(self._request.dbsession.query(self._DBGroup).filter(
            self._DBGroup.group_id.in_(group_ids)).order_by('group_id'))
        if not dbitems:
            return ''
        filename = '{0}.{1}.xml'.format(
            len(dbitems) == 1 and dbitems[0].group_id or
            make_id(self._request.registry['settings']['title'], 'token'),
            self._DBGroup.suffix)

        log_info(
            self._request, 'group_export',
            ' '.join([k.group_id for k in dbitems]))
        return db2web(self._request, dbitems, filename)

    # -------------------------------------------------------------------------
    def _save(self, dbgroup, profiles, values):
        """Save a user group.

        :type  dbgroup: .models.dbgroup.DBGroup
        :param dbgroup:
            Group to save.
        :param dict profiles:
            A dictionary such as ``{profile_id: (label, description),...}``.
        :param dict values:
            Form values.
        :rtype: :class:`~.models.dbgroup.DBGroup` instance or ``None``
        """
        creation = dbgroup is None
        dbgroup = dbgroup or self._DBGroup()

        # Update group
        record = {k: values[k] for k in values if k[:4] not in (
            'mbr:', 'shw:', 'pfl:')}
        if not creation:
            record['group_id'] = dbgroup.group_id
        error = dbgroup.record_format(record)
        if error:  # pragma: nocover
            self._request.session.flash(error, 'alert')
            return None
        for field in record:
            if getattr(dbgroup, field) != record[field]:
                setattr(dbgroup, field, record[field])

        # Save
        if creation:
            try:
                self._request.dbsession.add(dbgroup)
                self._request.dbsession.flush()
            except (IntegrityError, FlushError):
                self._request.session.flash(
                    _('This group already exists.'), 'alert')
                return None

        # Update users
        self._users_update(dbgroup)

        # Update profiles
        self._profiles_update(dbgroup, profiles, values)

        return dbgroup

    # -------------------------------------------------------------------------
    def _users_update(self, dbgroup):
        """Update the list of group users.

        :type  dbgroup: .models.dbwarhouse.DBGroup
        :param dbgroup:
            SQLAlchemy object for the current group.
        """
        if not self._request.has_permission('user-create'):
            return

        is_set = {}
        for value in self._request.POST:
            if value[:4] == 'mbr:':
                is_set[value[4:]] = True
            elif value[:4] == 'shw:' and value[4:] not in is_set:
                is_set[value[4:]] = False

        for user_id, user_set in is_set.items():
            dbgroup_user = self._request.dbsession.query(
                DBGroupUser).filter_by(
                    group_id=dbgroup.group_id,
                    user_id=int(user_id)).first()
            if dbgroup_user is not None and not user_set:
                self._request.dbsession.delete(dbgroup_user)
            elif dbgroup_user is None and user_set:
                self._request.dbsession.add(DBGroupUser(
                    group_id=dbgroup.group_id, user_id=int(user_id)))

    # -------------------------------------------------------------------------
    def _user_paging(self, action, dbgroup=None):
        """Return a paging object for users.

        :param str action:
            Current action.
        :type  dbgroup: .models.dbwarhouse.DBGroup
        :param dbgroup: (optional)
            If not ``None``, users are only users of the group.
        :rtype: tuple
        :return:
            A tuple such as ``(user_paging, defaults, user_filter)``.
        """
        # Filter
        paging_id = 'group_users'
        ufilter = Filter(
            self._request, paging_id, (
                ('login', _('Login'), False, None),
                ('last_name', _('Last name'), False, None),
                ('email', _('Email'), False, None),
                ('status', _('Status'), False,
                 [('', ' ')] + list(DBUser.status_labels.items()))),
            (('status', '=', 'active'),),
            remove=action[:4] == 'crm!' and action[4:] or None)

        # Database query
        defaults = Paging.params(
            self._request, paging_id, '+last_name', default_display='list')
        dbquery = ufilter.sql(
            self._request.dbsession.query(
                DBUser.user_id, DBUser.login,
                DBUser.first_name, DBUser.last_name,
                DBUser.honorific, DBUser.email,
                DBUser.email_hidden, DBUser.status,
                DBUser.last_login, DBUser.attachments_key,
                DBUser.picture), 'users')
        if dbgroup is not None:
            dbquery = dbquery.filter(DBUser.user_id.in_(
                [k.user_id for k in dbgroup.users]))
        oby = getattr(DBUser, defaults['sort'][1:])
        dbquery = dbquery.order_by(
            desc(oby) if defaults['sort'][0] == '-' else oby)

        return Paging(self._request, paging_id, dbquery, defaults), \
            dict(defaults), ufilter

    # -------------------------------------------------------------------------
    def _profiles_update(self, dbgroup, profiles, values):
        """Update the list of profiles.

        :type  dbgroup: .models.dbwarhouse.DBGroup
        :param dbgroup:
            SQLAlchemy object for the current group.
        :param dict profiles:
            A dictionary such as ``{profile_id: (label, description),...}``.
        :param dict values:
            Form values.
        """
        if not self._request.has_permission('user-create'):
            return

        group_profiles = {k.profile_id: k for k in dbgroup.profiles}
        for profile_id in profiles:
            value = values['pfl:{0}'.format(profile_id)]
            if value and profile_id not in group_profiles:
                self._request.dbsession.add(DBGroupProfile(
                    group_id=dbgroup.group_id, profile_id=profile_id))
            elif not value and profile_id in group_profiles:
                self._request.dbsession.delete(
                    self._request.dbsession.query(DBGroupProfile)
                    .filter_by(
                        group_id=dbgroup.group_id, profile_id=profile_id)
                    .first())



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``views.profile`` class."""

# pylint: disable = unused-import
from ..models.dbgroup import DBGroup  # noqa
# pylint: enable = unused-import
from . import DBUnitTestCase, FunctionalTestCase


# =============================================================================
class UViewsProfileProfileView(DBUnitTestCase):
    """Unit test class for testing :class:`views.profile.ProfileView`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from json import dumps
        from ..models.dbprofile import DBProfile, DBProfilePrincipal

        super(UViewsProfileProfileView, self).setUp()

        dbsession = self.request.dbsession

        record = {
            'profile_id': 'user_creator',
            'i18n_label': dumps({'en': 'User account manager'})}
        DBProfile.record_format(record)
        dbprofile = DBProfile(**record)
        dbprofile.principals.append(
            DBProfilePrincipal(principal='user.creator'))
        dbsession.add(dbprofile)

        record = {
            'profile_id': 'user_editor',
            'i18n_label': dumps({'en': 'User account editor'})}
        DBProfile.record_format(record)
        dbprofile = DBProfile(**record)
        dbprofile.principals.append(
            DBProfilePrincipal(principal='user.editor'))
        dbsession.add(dbprofile)

        self.configurator.add_route('home', '/')
        self.configurator.add_route('profile_index', '/profile/index')
        self.configurator.add_route('profile_create', '/profile/create')
        self.configurator.add_route(
            'profile_view', '/profile/view/{profile_id}')
        self.configurator.add_route(
            'profile_edit', '/profile/edit/{profile_id}')

    # -------------------------------------------------------------------------
    def test_index(self):
        """[u:views.profile.ProfileView.index]"""
        from collections import namedtuple
        from ..lib.filter import Filter
        from ..views.profile import ProfileView
        from . import DummyPOST

        self.request.matched_route = namedtuple('Route', 'name')(
            name='profile_index')

        response = ProfileView(self.request).index()
        self.assertIn('pfilter', response)
        self.assertIsInstance(response['pfilter'], Filter)
        self.assertTrue(response['pfilter'].is_empty())
        self.assertIn('paging', response)
        self.assertIn('i_creator', response)
        self.assertIn('i_editor', response)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Profiles')

        self.request.is_xhr = True
        self.request.POST = DummyPOST()
        response = ProfileView(self.request).index()
        self.assertIsInstance(response, dict)
        self.assertFalse(response)

    # -------------------------------------------------------------------------
    def test_index_delete(self):
        """[u:views.profile.ProfileView.index] delete profile"""
        from ..models.dbprofile import DBProfile
        from ..views.profile import ProfileView

        dbsession = self.request.dbsession
        self.assertEqual(dbsession.query(DBProfile).count(), 2)

        self.request.POST = {'del!user_creator.x': True}
        ProfileView(self.request).index()
        self.assertEqual(dbsession.query(DBProfile).count(), 1)

    # -------------------------------------------------------------------------
    def test_index_import(self):
        """[u:views.profile.ProfileView.index] import profile"""
        from cgi import FieldStorage
        from . import TEST1_PFL_XML, DummyPOST
        from ..views.profile import ProfileView

        self.request.POST = DummyPOST({'imp!.x': True})
        response = ProfileView(self.request).index()
        self.assertIn('paging', response)
        self.assertNotIn(
            'user_viewer', [k.profile_id for k in response['paging']])

        with open(TEST1_PFL_XML, 'r') as hdl:
            input_file = FieldStorage()
            input_file.file = hdl
            input_file.filename = TEST1_PFL_XML
            self.request.POST.multikeys = {'file': (input_file,)}
            response = ProfileView(self.request).index()
        self.assertIn('paging', response)
        self.assertIn(
            'user_viewer', [k.profile_id for k in response['paging']])

    # -------------------------------------------------------------------------
    def test_index_export(self):
        """[u:views.profile.ProfileView.index] export profile"""
        from pyramid.response import Response
        from ..views.profile import ProfileView

        self.request.POST = {'exp!foo.x': True}
        response = ProfileView(self.request).index()
        self.assertIsInstance(response, dict)

        self.request.POST = {'exp!user_creator.x': True}
        response = ProfileView(self.request).index()
        self.assertIsInstance(response, Response)
        self.assertEqual(response.content_type, 'application/xml')
        self.assertIn(b'user_creator', response.body)

    # -------------------------------------------------------------------------
    def test_index_filter(self):
        """[u:views.profile.ProfileView.index_filter]"""
        from ..views.profile import ProfileView

        self.request.is_xhr = True
        self.request.params = {'field': 'i18n_label', 'value': 'user'}
        data = ProfileView(self.request).index_filter()
        self.assertIsInstance(data, list)
        self.assertIn('User account manager', data)

    # -------------------------------------------------------------------------
    def test_view(self):
        """[u:views.profile.ProfileView.view]"""
        from collections import namedtuple
        from pyramid.httpexceptions import HTTPNotFound
        from ..views.profile import ProfileView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='profile_view')

        self.request.matchdict = {'profile_id': 'foo'}
        self.assertRaises(HTTPNotFound, ProfileView(self.request).view)

        self.request.matchdict = {'profile_id': 'user_creator'}
        response = ProfileView(self.request).view()
        self.assertIn('form', response)
        self.assertEqual(
            self.request.breadcrumbs.current_title(),
            'Profile "User account manager"')

    # -------------------------------------------------------------------------
    def test_view_export(self):
        """[u:views.profile.ProfileView.view] export profiles"""
        from pyramid.response import Response
        from ..views.profile import ProfileView

        self.request.POST = {'exp!.x': True}
        self.request.matchdict = {'profile_id': 'user_creator'}

        response = ProfileView(self.request).view()
        self.assertIsInstance(response, Response)
        self.assertEqual(response.content_type, 'application/xml')
        self.assertIn(b'user_creator', response.body)

    # -------------------------------------------------------------------------
    def test_create(self):
        """[u:views.profile.ProfileView.edit] creation"""
        from collections import namedtuple
        from ..views.profile import ProfileView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='profile_create')

        response = ProfileView(self.request).edit()
        self.assertIn('form', response)
        self.assertIn('label', response)
        self.assertIsNone(response['label'])
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Profile Creation')

    # -------------------------------------------------------------------------
    def test_create_save(self):
        """[u:views.profile.ProfileView.edit] save creation"""
        from collections import namedtuple
        from ..views.profile import ProfileView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='profile_create')
        self.request.POST = {
            'profile_id': 'user_editor',
            'label_en': 'User account observer',
            'description_en': 'These users can view user accounts.',
            'sav!.x': True}

        # Existing profile
        response = ProfileView(self.request).edit()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('exists', errors[0])
        self.assertIn('form', response)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Profile Creation')

        # New profile, missing label
        self.request.POST['profile_id'] = 'user_viewer'
        self.request.POST['label_en'] = ''
        ProfileView(self.request).edit()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('Correct', errors[0])
        # pylint: disable = no-member
        self.request.dbsession.rollback()
        # pylint: enable = no-member

        # New profile OK
        self.request.POST['label_en'] = 'User account observer'
        ProfileView(self.request).edit()
        self.assertFalse(self.request.session.pop_flash('alert'))

    # -------------------------------------------------------------------------
    def test_edit(self):
        """[u:views.profile.ProfileView.edit]"""
        from collections import namedtuple
        from ..models.dbprofile import DBProfile
        from ..views.profile import ProfileView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='profile_edit')
        self.request.matchdict = {'profile_id': 'user_editor'}

        response = ProfileView(self.request).edit()
        self.assertIn('form', response)
        self.assertIn('label', response)
        self.assertEqual(response['label'], 'User account editor')
        self.assertIn('dbprofile', response)
        self.assertIsInstance(response['dbprofile'], DBProfile)
        self.assertEqual(
            self.request.breadcrumbs.current_title(),
            'Profile "User account editor" Edition')

    # -------------------------------------------------------------------------
    def test_edit_save(self):
        """[u:views.profile.ProfileView.edit] save"""
        from pyramid.httpexceptions import HTTPFound
        from ..security import PRINCIPALS
        from ..views.profile import ProfileView

        self.request.matchdict = {'profile_id': 'user_editor'}
        self.request.registry['principals'] = list(PRINCIPALS)
        self.request.POST = {
            'description_en': 'These users can edit and view user accounts.',
            'pcpl:user.viewer': True, 'sav!.x': True}

        ProfileView(self.request).edit()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('Correct', errors[0])

        self.request.POST['label_en'] = 'User account editor'
        self.assertIsInstance(ProfileView(self.request).edit(), HTTPFound)


# =============================================================================
class FViewsProfileProfileView(FunctionalTestCase):
    """Functional test class for testing :class:`views.profile.ProfileView`."""

    # -------------------------------------------------------------------------
    def test_index(self):
        """[f:views.profile.ProfileView.index]"""
        self.login('test1')
        self.testapp.get('/profile/index', status=403)

        self.login('admin')
        response = self.testapp.get('/profile/index', status=200)
        self.assertIn(b'Profiles', response.body)
        self.assertIn(b'Create', response.body)
        self.assertIn(b'Import', response.body)
        self.assertIn(b'Export profile', response.body)

    # -------------------------------------------------------------------------
    def test_view(self):
        """[f:views.profile.ProfileView.view]"""
        self.login('test1')
        self.testapp.get('/profile/view/user_editor', status=403)

        self.login('admin')
        response = self.testapp.get('/profile/view/user_editor', status=200)
        self.assertIn(b'Edit', response.body)

    # -------------------------------------------------------------------------
    def test_create(self):
        """[f:views.profile.ProfileView.create]"""
        self.login('test1')
        self.testapp.get('/profile/create', status=403)

        self.login('admin')
        response = self.testapp.get('/profile/create', status=200)
        self.assertIn(b'Create', response.body)

    # -------------------------------------------------------------------------
    def test_edit(self):
        """[f:views.profile.ProfileView.edit]"""
        self.login('test1')
        self.testapp.get('/profile/edit/user_editor', status=403)

        self.login('admin')
        response = self.testapp.get('/profile/edit/user_editor', status=200)
        self.assertIn(b'Save', response.body)



"""Main class to initialize the application."""

from sys import exit as sys_exit
from os import makedirs, chdir
from os.path import exists, dirname
from shutil import rmtree
from configparser import ConfigParser
from pkg_resources import get_distribution

import transaction

from pyramid.csrf import SessionCSRFStoragePolicy

from .lib.i18n import _, add_translation_dirs, translate
from .lib.config import config_get, settings_get_list
from .lib.log import log_activity_setup
from .lib.utils import tounicode, check_chrysalio_js, check_chrysalio_css
from .lib.xml import check_chrysalio_rng
from .includes.themes import create_default_theme
from .models import get_tm_dbsession
from .models.dbsettings import DBSettings


# =============================================================================
class Initialize(object):
    """Initialization application class.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """

    # -------------------------------------------------------------------------
    def __init__(self, configurator):
        """Constructor method."""
        self._configurator = configurator

    # -------------------------------------------------------------------------
    def complete(self, global_config, package, populate_script):
        """Check settings and create registry objects.

        :param dict global_config:
            Dictionary describing the INI file with keys ``__file__`` and
            ``here``.
        :param str package:
            Name of the calling package.
        :param str populate_script:
            Name of populate script.
        """
        # Translations
        add_translation_dirs(self._configurator, package)

        # Version
        self._configurator.registry['version'] = '{0} {1}'.format(
            get_distribution(package).project_name,
            get_distribution(package).version)

        # CSRF policy
        self._configurator.set_csrf_storage_policy(
            SessionCSRFStoragePolicy())
        self._configurator.set_default_csrf_options(require_csrf=True)

        # General settings
        chdir(dirname(global_config['__file__']))
        settings = self._configurator.get_settings()
        settings['__file__'] = global_config['__file__']
        if not settings.get('site.uid'):
            sys_exit(translate(_('*** You must define a Web site ID.')))
        config = ConfigParser({'here': global_config['here']})
        config.read(tounicode(global_config['__file__']), encoding='utf8')
        default_email = config_get(config, 'Populate', 'admin.email')
        if not default_email:
            sys_exit(translate(_('*** The administrator must have an email.')))
        with transaction.manager:
            dbsession = get_tm_dbsession(
                self._configurator.registry['dbsession_factory'],
                transaction.manager)
            if 'testing' not in settings and not DBSettings.exists(dbsession):
                sys_exit(translate(
                    _('*** Run "${s}" script!', {'s': populate_script})))
            self._configurator.registry['settings'] = DBSettings.db2dict(
                settings, dbsession, default_email)

        # Theme management
        includes = settings_get_list(settings, 'chrysalio.includes')
        if 'chrysalio.includes.themes' not in includes:
            create_default_theme(self._configurator, package)

        # Create log directory
        self._configurator.registry['log_activity'] = \
            log_activity_setup(config)

        # Temporary directory
        if settings.get('temporary'):
            if exists(settings['temporary']):
                rmtree(settings['temporary'])
            makedirs(settings['temporary'])

        # Check Relax NG, Js and Css
        if package != 'chrysalio':  # pragma: nocover
            check_chrysalio_rng('{0}:RelaxNG'.format(package))
            check_chrysalio_css('{0}:Static/Css'.format(package))
            check_chrysalio_js('{0}:Static/Js'.format(package))

    # -------------------------------------------------------------------------
    def add_static_views(self, package, statics):
        """Add static views.

        :param str package:
            Name of the calling package.
        :param tuple statics:
            A tuple of tuples such as ``(static_name, static_abs_path)``.
        """
        for theme in self._configurator.registry['themes']:
            theme = '/theme/{0}'.format(theme.lower()) if theme else ''
            for static in statics:
                url = '{0}/{1}/{2}/'.format(theme, package, static[0])
                if self._configurator.introspector.get(
                        'static views', url) is None:
                    self._configurator.add_static_view(url, static[1])



"""Various functions to parse configuration files or settings."""

try:  # pragma: nocover
    from os import scandir
except ImportError:  # pragma: nocover
    from scandir import scandir
from os.path import exists, join, abspath
from sys import version_info
from fnmatch import fnmatch

from pyramid.asset import abspath_from_asset_spec
from pyramid.security import Allow


# =============================================================================
def update_acl(configurator, principals):
    """Update root factory Access Control List (ACL) with permissions.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    :param tuple principals:

    ``principals`` is a tuple such as::

        DOCUMENT_PRINCIPALS = (
            ('document', _('Document management'), (
                ('viewer', _('View all documents'), (
                    'document-view',)),
                ('editor', _('Edit any document'), (
                    'document-view', 'document-edit'))
            )),
        )
    """
    root_factory = configurator.introspector.get(
        'root factories', None).get('factory')
    root_factory.__acl__ += [
        (Allow, '{0}.{1}'.format(group[0], principal[0]), principal[2])
        for group in principals for principal in group[2]]


# =============================================================================
def config_get(config, section, option, default=None):
    """Retrieve a value from a configuration object.

    :type  config: configparser.ConfigParser
    :param config:
        Configuration object.
    :param str section:
        Section name.
    :param str option:
        Option name.
    :param str default: (optional)
        Default value
    :rtype: str
        Read value or default value.
    """
    if not config.has_option(section, option):
        return default
    value = config.get(section, option)
    if version_info > (3, 0):
        return value or None  # pragma: no cover
    return value.decode('utf8') if isinstance(value, str) \
        else value  # pragma: no cover


# =============================================================================
def config_get_list(config, section, option, default=None):
    """Retrieve a list of values from a configuration object.

    :type  config: configparser.ConfigParser
    :param config:
        Configuration object.
    :param str section:
        Section name.
    :param str option:
        Option name.
    :param list default: (optional)
        Default values.
    :rtype: list
    """
    if not config.has_option(section, option):
        return default or []
    values = config_get(config, section, option)
    return [k.strip() for k in values.split(',') if k] if values else []


# =============================================================================
def config_get_namespace(config, section, namespace):
    """Retrieve all options beginning by a name space.

    :type  config: configparser.ConfigParser
    :param config:
        Configuration object.
    :param str section:
        Section name.
    :param str namespace:
        Prefix of options to retrieve.
    :rtype: dict
    """
    values = {}
    ns_len = len(namespace) + 1
    if not config.has_section(section):
        return values

    for option in config.options(section):
        if option.startswith('{0}.'.format(namespace)):
            values[option[ns_len:].replace('.', '_')] = config_get(
                config, section, option)
    return values


# =============================================================================
def settings_get_list(settings, option, default=None):
    """Retrieve a list of values from a settings dictionary.

    :type  settings: pyramid.registry.Registry.settings
    :param settings:
        Settings object.
    :param str option:
        Option name.
    :param list default: (optional)
        Default values.
    :rtype: list
    """
    if not settings.get(option):
        return default or []
    return [k.strip() for k in settings[option].split(',') if k]


# =============================================================================
def settings_get_namespace(settings, namespace):
    """Retrieve all options beginning by a name space.

    :type  settings: pyramid.registry.Registry.settings
    :param settings:
        Settings object.
    :param str namespace:
        Prefix of options to retrieve.
    :rtype: dict
    """
    values = {}
    ns_len = len(namespace) + 1

    for option in settings:
        if option.startswith('{0}.'.format(namespace)):
            values[option[ns_len:].replace('.', '_')] = settings[option]
    return values


# =============================================================================
def settings_get_directories(settings, namespace, conf_file):
    """Retrieve all directories whose root is contained in one of the
    directories listed in ``namespace.roots``, name matches one of the patterns
    listed in ``namespace.patterns`` and containing the file ``filename``.

    :type  settings: pyramid.registry.Registry.settings
    :param settings:
        Settings object.
    :param str namespace:
        Prefix of options to retrieve.
    :param str namespace:
        Prefix of options to retrieve.
    :param str conf_file:
        Name of configuration file to search in each directory.
    :rtype: dict
    """
    directories = {}
    done = set()
    patterns = settings_get_list(
        settings, '{0}.patterns'.format(namespace), '')
    for root in settings_get_list(
            settings, '{0}.roots'.format(namespace), ''):
        root = abspath(abspath_from_asset_spec(root))
        if root in done or not exists(root):
            continue
        for entry in scandir(root):
            if entry.is_dir() and exists(join(entry.path, conf_file)):
                for pattern in patterns:
                    if fnmatch(entry.name, pattern):
                        directories[entry.name] = entry.path
                        break

    return directories



"""Theme management."""

from sys import exit as sys_exit
from os.path import exists, join, abspath
from configparser import ConfigParser

from pyramid.asset import abspath_from_asset_spec
from pyramid.config import Configurator

from ..lib.i18n import _, translate
from ..lib.config import config_get_namespace
from ..lib.config import settings_get_list, settings_get_directories


# =============================================================================
def includeme(configurator):
    """Function to include themes.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    if isinstance(configurator, Configurator):
        load_themes(configurator)


# =============================================================================
def load_themes(configurator):
    """Load available themes in registry.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.

    Set into registry a tuple such as ``(default_theme, theme_dict)`` where
    ``theme_dict`` is a dictionary like ``{'theme_id': theme_info,...}``.
    ``theme_info`` is a dictionary with keys ``path``, ``layout``,
    ``login``,...
    """
    # Read list of available themes
    themes = {}
    directories = settings_get_directories(
        configurator.get_settings(), 'theme', 'theme.conf')
    for theme_id, dir_path in directories.items():
        config = ConfigParser({'here': dir_path})
        config.read(join(dir_path, 'theme.conf'), encoding='utf8')
        themes[theme_id] = {
            'path': dir_path,
            'name': config_get_namespace(config, 'Description', 'name')}
        for section in ('Templates', 'Static'):
            if not config.has_section(section):
                sys_exit(translate(_(
                    '*** Theme "${t}" must have a section [${s}].',
                    {'t': theme_id, 's': section})))
            if section == 'Templates' and \
                    not config.has_option(section, 'layout'):
                sys_exit(translate(_(
                    '*** An option is missing for [${s}] of theme "${t}".',
                    {'s': section, 't': theme_id})))
            if section == 'Static' and (
                    not config.has_option(section, 'css') or
                    not config.has_option(section, 'js')):
                sys_exit(translate(_(
                    '*** An option is missing for [${s}] of theme "${t}".',
                    {'s': section, 't': theme_id})))
            for item in config.items(section):
                if item[0] == 'here':
                    continue
                if not item[1]:
                    sys_exit(translate(_(
                        '*** Option "${o}" of theme "${t}" must be defined.',
                        {'o': item[0], 't': theme_id})))
                path = abspath(abspath_from_asset_spec(item[1]))
                if not exists(path):
                    sys_exit(translate(_(
                        '*** File "${f}" of theme "${t}" does not exist.',
                        {'f': path, 't': theme_id})))
                if section == 'Templates':
                    themes[theme_id][item[0]] = path
                else:
                    configurator.add_static_view(
                        '/theme/{theme_id}/{static_id}/'.format(
                            theme_id=theme_id.lower(), static_id=item[0]),
                        path)
    configurator.registry['themes'] = themes
    configurator.commit()

    # Read default theme
    default = configurator.registry['settings']['theme']
    default = settings_get_list(
        configurator.get_settings(), 'theme.patterns', [''])[0] \
        if not default and default not in themes and themes else default
    if default not in themes:
        sys_exit(translate(_('*** Must have an existing default theme.')))
    configurator.registry['settings']['theme'] = default


# =============================================================================
def create_default_theme(configurator, package):
    """If the theme management is not activated, it creates a default theme in
    registry.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    :param str package:
        Name of the calling package.
    """
    # Templates
    settings = configurator.get_settings()
    path = abspath(abspath_from_asset_spec('{0}:Templates'.format(package)))
    configurator.registry['themes'] = {
        '': {'path': None, 'name': {},
             'layout': settings.get('site.layout', join(path, 'layout.pt'))}}

    # Static
    path = abspath(abspath_from_asset_spec('{0}:Static'.format(package)))

    configurator.add_static_view(
        '/css/', settings.get('site.css', join(path, 'Css')))
    configurator.add_static_view(
        '/js/', settings.get('site.js', join(path, 'Js')))
    configurator.add_static_view(
        '/images/', settings.get('site.images', join(path, 'Images')))
    configurator.add_static_view(
        '/custom/', settings.get('site.custom', join(path, 'Custom')))

    configurator.commit()


# =============================================================================
def theme_template(request, template_id):
    """Retrieve the absolute file path of the file ``template_id`` of the
    current theme.

    :param request: pyramid.request.Request`
    :param request:
        Current request.
    :param str template_id:
        ID of the file to retrieve.
    :rtype: str
    :return:
        Absolute path to the template file or ``None``.
    """
    theme_id = request.session.get(
        'theme', request.registry['settings']['theme'])
    return request.registry['themes'][theme_id].get(template_id)


# =============================================================================
def theme_static_prefix(request):
    """Return the URL prefix for the current theme.

    :param request: pyramid.request.Request`
    :param request:
        Current request.
    :rtype: str
        A string such as ``/theme/alternative`` for theme `Alternative` or ''.
    """
    if request is None:
        return ''
    theme_id = request.session.get(
        'theme', request.registry['settings']['theme'])
    return '/theme/{0}'.format(theme_id.lower()) if theme_id else ''


# =============================================================================
def theme_has_static(request, static_id):
    """Return ``True`` if the descriminator exists in static views.

    :param request: pyramid.request.Request`
    :param request:
        Current request.
    :param str static_id:
        ID of static view to check.
    :rtype: bool
    """
    theme_id = request.session.get(
        'theme', request.registry['settings']['theme'])
    return request.registry.introspector.get(
        'static views', '/theme/{theme_id}/{static_id}/'.format(
            theme_id=theme_id.lower(), static_id=static_id)) is not None



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``models.dbuser`` classes."""

from datetime import datetime, date, timedelta
from json import dumps

from lxml import etree
from colander import SchemaNode

from pyramid.i18n import TranslationString

from ..models.dbuser import DBUser
from ..models.dbgroup import DBGroup
from . import DBUnitTestCase


# =============================================================================
class DummyAuthority(object):
    """Class to simulate an authority for authentification as LDAP."""

    # -------------------------------------------------------------------------
    def __init__(self, settings):
        """Constructor method."""

    # -------------------------------------------------------------------------
    @classmethod
    def get(cls, ignored_, login, password, dbuser_class):
        """Get user from dummy authority."""
        if login == 'test100' and password == 'test100pwd':
            dbuser = dbuser_class(
                login=login, password=password, last_name='HONNÊTE',
                email='test100@chrysal.io')
            return dbuser, None
        return None, 'Dummy: ID or password is incorrect.'

    # -------------------------------------------------------------------------
    @classmethod
    def check(cls, ignored_, login, password, unused_):
        """Check user authorization according to dummy authority."""
        if login == 'test101' and password == 'test101pwd':
            return None
        return 'Dummy: ID or password is incorrect.'


# =============================================================================
class UModelsDBUserDBUser(DBUnitTestCase):
    """Unit test class for :class:`models.dbuser.DBUser`."""

    # -------------------------------------------------------------------------
    @classmethod
    def _make_one(cls):
        """Make an ``DBUser`` object."""
        dbuser = DBUser(
            login='test1', status='active', first_name='Édith',
            last_name='AVULEUR', honorific='Mrs', email='test1@chrysal.io',
            theme='Default', page_size=80, account_creation=datetime.now(),
            account_update=datetime.now(), expiration=datetime(2100, 1, 1),
            last_login=datetime.now())
        dbuser.set_password('test1pwd')
        return dbuser

    # -------------------------------------------------------------------------
    def test_set_password(self):
        """[u:models.dbuser.DBUser.set_password]"""
        user = DBUser(
            login='test1', first_name='Édith', last_name='AVULEUR',
            honorific='Mrs', email='test1@chrysal.io')
        user.set_password('test1pwd')
        self.assertTrue(user.password.startswith('$'))
        user.set_password(
            '$2b$12$c67DI/dlDjSqqYtGcfmhKeXOp/6ohidEFUE3/nsOeBMnmWmuQYGwa')
        self.assertEqual(
            user.password,
            '$2b$12$c67DI/dlDjSqqYtGcfmhKeXOp/6ohidEFUE3/nsOeBMnmWmuQYGwa')

    # -------------------------------------------------------------------------
    def test_check_password(self):
        """[u:models.dbuser.DBUser.check_password]"""
        dbuser = self._make_one()
        self.assertFalse(dbuser.check_password(None))
        self.assertFalse(dbuser.check_password('foo'))
        self.assertTrue(dbuser.check_password('test1pwd'))
        self.assertTrue(
            dbuser.password_update > datetime.now() - timedelta(seconds=2))

    # -------------------------------------------------------------------------
    def test_get(self):
        """[u:models.dbuser.DBUser.get]"""
        dbsession = self.request.dbsession

        # No login
        dbuser, error = DBUser.get(self.request)
        self.assertIsNone(dbuser)
        self.assertEqual(error, 'ID or password is incorrect.')

        # Unknown user
        dbuser, error = DBUser.get(self.request, 'test1')
        self.assertIsNone(dbuser)
        self.assertEqual(error, 'ID or password is incorrect.')

        # Invalid password
        self.add_user({
            'login': 'test1', 'first_name': 'Édith', 'last_name': 'AVULEUR',
            'password': 'test1pwd', 'email': 'test1@chrysal.io'})
        dbuser, error = DBUser.get(self.request, 'test1', 'foo')
        self.assertIsNone(dbuser)
        self.assertEqual(error, 'ID or password is incorrect.')

        # Connection ok
        dbuser, error = DBUser.get(self.request, 'test1', 'test1pwd')
        self.assertIsInstance(dbuser, DBUser)
        self.assertIsNone(error)
        user = dbsession.query(DBUser).filter_by(login='test1').first()
        self.assertTrue(
            user.last_login > datetime.now() - timedelta(seconds=2))

        # Inactive account
        self.add_user({
            'login': 'test2', 'first_name': 'Sophie', 'last_name': 'FONFEC',
            'status': 'inactive', 'password': 'test2pwd',
            'email': 'test2@chrysal.io'})
        dbuser, error = DBUser.get(self.request, 'test2', 'test2pwd')
        self.assertIsNone(dbuser)
        self.assertEqual(error, 'Your account is not active.')

        # Locked account
        self.add_user({
            'login': 'test3', 'first_name': 'Guy', 'last_name': 'LIGUILI',
            'status': 'locked', 'password': 'test3pwd',
            'email': 'test3@chrysal.io'})
        dbuser, error = DBUser.get(self.request, 'test3', 'test3pwd')
        self.assertIsNone(dbuser)
        self.assertEqual(error, 'Your account is locked.')

        # Expired account
        self.add_user({
            'login': 'test4', 'first_name': 'Gédéon',
            'last_name': 'TEUSEMANIE', 'password': 'test4pwd',
            'email': 'test4@chrysal.io', 'expiration': '2018-01-01'})
        dbuser, error = DBUser.get(self.request, 'test4', 'test4pwd')
        self.assertIsNone(dbuser)
        self.assertEqual(error, 'Your account has expired.')

        # With authority: creation with invalid password
        self.request.registry['authorities'] = {'foo': DummyAuthority({})}
        dbuser, error = DBUser.get(self.request, 'test100', 'foo')
        self.assertIsNone(dbuser)
        self.assertEqual(error, 'Dummy: ID or password is incorrect.')

        # With authority: creation ok
        dbuser, error = DBUser.get(self.request, 'test100', 'test100pwd')
        self.assertIsNotNone(dbuser)
        self.assertIsNone(error)

        # With authority: check with authority not available
        self.add_user({
            'login': 'test101', 'first_name': 'Bob',
            'last_name': 'INDEFIL', 'password': 'test101pwd',
            'email': 'test101@chrysal.io', 'authority': 'dummy'})
        dbuser, error = DBUser.get(self.request, 'test101', 'test101pwd')
        self.assertIsNone(dbuser)
        self.assertEqual(error, 'Authority not available.')

        # With authority: check with invalid password
        self.request.registry['authorities'] = {'dummy': DummyAuthority({})}
        dbuser, error = DBUser.get(self.request, 'test101', 'foo')
        self.assertIsNone(dbuser)
        self.assertEqual(error, 'Dummy: ID or password is incorrect.')

        # With authority: check ok
        dbuser, error = DBUser.get(self.request, 'test101', 'test101pwd')
        self.assertIsNotNone(dbuser)
        self.assertIsNone(error)

    # -------------------------------------------------------------------------
    def test_set_session(self):
        """[u:models.dbuser.DBUser.set_session]"""
        # pylint: disable = too-many-statements
        from ..models.dbprofile import DBProfile, DBProfilePrincipal
        from ..models.dbuser import DBUserProfile
        from ..models.dbgroup import DBGroupProfile

        dbsession = self.request.dbsession
        dbuser = self._make_one()
        dbsession.add(dbuser)
        self.request.registry.settings['languages'] = 'en, fr'

        # Check language
        dbuser.set_session(self.request)
        self.assertNotIn('lang', self.request.session)
        dbuser.language = 'fr'
        dbuser.set_session(self.request)
        self.assertEqual(self.request.session['lang'], 'fr')
        dbuser.language = 'es'
        dbuser.set_session(self.request)
        self.assertEqual(self.request.session['lang'], 'en')

        # Check theme
        self.assertIn('theme', self.request.session)
        self.assertEqual(self.request.session['theme'], '')
        self.request.registry['themes'] = {'Default': {}, 'Alternative': {}}
        self.request.registry['settings']['theme'] = 'Default'
        dbuser.theme = 'Foo'
        dbuser.set_session(self.request)
        self.assertEqual(self.request.session['theme'], 'Default')
        dbuser.theme = 'Alternative'
        dbuser.set_session(self.request)
        self.assertEqual(self.request.session['theme'], 'Alternative')

        # Check general information
        self.assertIn('user', self.request.session)
        user_session = self.request.session['user']
        self.assertIn('user_id', user_session)
        self.assertIn('login', user_session)
        self.assertIn('name', user_session)
        self.assertIn('principals', user_session)
        self.assertEqual(user_session['login'], 'test1')
        self.assertEqual(user_session['name'], 'Édith AVULEUR')

        # Check groups and profiles
        dbprofile = DBProfile(
            profile_id='profile_manager',
            i18n_label=dumps({'en': 'Profile manager'}))
        dbprofile.principals.append(
            DBProfilePrincipal(principal='profile.creator'))
        dbsession.add(dbprofile)
        dbprofile = DBProfile(
            profile_id='user_manager',
            i18n_label=dumps({'en': 'User manager'}))
        dbprofile.principals.append(
            DBProfilePrincipal(principal='user.creator'))
        dbsession.add(dbprofile)

        dbgroup = DBGroup(
            group_id='managers', i18n_label=dumps({'en': 'Managers'}))
        dbsession.add(dbgroup)
        dbsession.flush()
        dbsession.add(DBUserProfile(
            user_id=dbuser.user_id, profile_id='user_manager'))
        dbsession.add(DBGroupProfile(
            group_id=dbgroup.group_id, profile_id='profile_manager'))
        dbuser.groups.append(dbgroup)
        dbuser.set_session(self.request)
        user_session = self.request.session['user']

        self.assertEqual(len(user_session['principals']), 2)
        self.assertIn('profile.creator', user_session['principals'])
        self.assertIn('user.creator', user_session['principals'])
        self.assertEqual(len(user_session['groups']), 1)
        self.assertEqual(user_session['groups'][0], 'managers')

        # Check administrator and modes
        self.request.session['menu'] = ()
        self.request.session['modes'] = ['home', 'Home', None]
        dbuser.status = 'administrator'
        dbuser.set_session(self.request)
        user_session = self.request.session['user']
        self.assertEqual(len(user_session['principals']), 1)
        self.assertEqual(user_session['principals'][0], 'system.administrator')
        self.assertNotIn('menu', self.request.session)

    # -------------------------------------------------------------------------
    def test_load_administrator(self):
        """[u:models.dbuser.DBUser.load_administrator]"""
        dbsession = self.request.dbsession
        record = {
            'password': 'adminpwd', 'last_name': 'Administrateur',
            'language': 'fr', 'time_zone': 'Europe/Paris'}

        error = DBUser.load_administrator(dbsession, record)
        self.assertIsInstance(error, TranslationString)

        self.add_user({
            'login': 'test1', 'first_name': 'Édith', 'last_name': 'AVULEUR',
            'password': 'test1pwd', 'email': 'test1@chrysal.io'})
        record['login'] = 'test1'
        error = DBUser.load_administrator(dbsession, record)
        self.assertIsNone(error)

        record['login'] = 'admin'
        error = DBUser.load_administrator(dbsession, record)
        self.assertIsInstance(error, TranslationString)

        record['email'] = 'admin@chrysal.io'
        error = DBUser.load_administrator(dbsession, record)
        self.assertIsNone(error)

        user = dbsession.query(DBUser).filter_by(login='admin').first()
        self.assertIsInstance(user, DBUser)
        self.assertTrue(user.password.startswith('$'))
        self.assertEqual(user.status, 'administrator')
        self.assertEqual(user.last_name, 'Administrateur')
        self.assertTrue(user.email_hidden)
        self.assertEqual(user.email, 'admin@chrysal.io')
        self.assertEqual(user.language, 'fr')
        self.assertEqual(user.time_zone, 'Europe/Paris')

    # -------------------------------------------------------------------------
    def test_xml2db(self):
        """[u:models.dbuser.DBUser.xml2db]"""
        dbsession = self.request.dbsession
        self.add_user({
            'login': 'test2', 'first_name': 'Sophie', 'last_name': 'FONFEC',
            'password': 'test2pwd', 'email': 'test2@chrysal.io'})

        user_elt = etree.XML(
            '<user>'
            '  <login>test2</login>'
            '  <password>test2pwd</password>'
            '  <firstname>Sophie</firstname>'
            '  <lastname>FONFEC</lastname>'
            '  <email>test2@chrysal.io</email>'
            '</user>')
        error = DBUser.xml2db(dbsession, user_elt)
        self.assertIsInstance(error, TranslationString)
        error = DBUser.xml2db(dbsession, user_elt, error_if_exists=False)
        self.assertIsNone(error)

        user_elt = etree.XML('<user><login>test1</login></user>')
        error = DBUser.xml2db(dbsession, user_elt)
        self.assertIsInstance(error, TranslationString)

        user_elt = etree.XML(
            '<user>'
            '  <login>test1</login>'
            '  <password must-change="true">test1pwd</password>'
            '  <firstname>Édith</firstname>'
            '  <lastname>AVULEUR</lastname>'
            '  <honorific>Mrs</honorific>'
            '  <email hidden="true">test1@chrysal.io</email>'
            '  <language>fr</language>'
            '  <theme>Alternative</theme>'
            '</user>')
        error = DBUser.xml2db(dbsession, user_elt)
        self.assertIsNone(error)
        user = dbsession.query(DBUser).filter_by(login='test1').first()
        self.assertIsInstance(user, DBUser)
        self.assertTrue(user.password_mustchange)
        self.assertEqual(user.first_name, 'Édith')
        self.assertEqual(user.last_name, 'AVULEUR')
        self.assertEqual(user.honorific, 'Mrs')
        self.assertEqual(user.email, 'test1@chrysal.io')
        self.assertTrue(user.email_hidden)
        self.assertEqual(user.language, 'fr')
        self.assertEqual(user.theme, 'Alternative')

    # -------------------------------------------------------------------------
    def test_xml2db_extra(self):
        """[u:models.dbuser.DBUser.xml2db_extra]"""
        from ..models.dbprofile import DBProfile

        dbsession = self.request.dbsession
        dbsession.add(DBProfile(
            profile_id='user_manager',
            i18n_label=dumps({'en': 'User manager'})))
        self.add_user({
            'login': 'test1', 'first_name': 'Édith', 'last_name': 'AVULEUR',
            'password': 'test1pwd', 'email': 'test1@chrysal.io'})
        dbuser = dbsession.query(DBUser).filter_by(login='test1').first()
        user_elt = etree.XML(
            '<user>'
            '  <login>test1</login>'
            '  <password>test1pwd</password>'
            '  <firstname>Édith</firstname>'
            '  <lastname>AVULEUR</lastname>'
            '  <email>test1@chrysal.io</email>'
            '  <profiles>'
            '    <profile>profile_manager</profile>'
            '    <profile>user_manager</profile>'
            '  </profiles>'
            '</user>')
        dbuser.xml2db_extra(
            dbsession, user_elt, {'profiles': ['user_manager']})
        self.assertEqual(len(dbuser.profiles), 1)

    # -------------------------------------------------------------------------
    def test_record_from_xml(self):
        """[u:models.dbuser.DBUser.record_from_xml]"""
        user_elt = etree.XML(
            '<user>'
            '  <login>test1</login>'
            '  <password must-change="true">test1pwd</password>'
            '  <firstname>Édith</firstname>'
            '  <lastname>AVULEUR</lastname>'
            '  <honorific>Mrs</honorific>'
            '  <email hidden="true">test1@chrysal.io</email>'
            '  <expiration>2018-01-01</expiration>'
            '  <profiles>'
            '    <profile>profile_manager</profile>'
            '    <profile>user_manager</profile>'
            '  </profiles>'
            '</user>')
        record = DBUser.record_from_xml('test1', user_elt)
        self.assertEqual(record['status'], 'active')
        self.assertEqual(record['login'], 'test1')
        self.assertTrue(record['password_mustchange'])
        self.assertEqual(record['first_name'], 'Édith')
        self.assertEqual(record['last_name'], 'AVULEUR')
        self.assertEqual(record['honorific'], 'Mrs')
        self.assertIsNone(record['language'])
        self.assertEqual(record['expiration'], '2018-01-01')

    # -------------------------------------------------------------------------
    def test_record_format(self):
        """[u:models.dbuser.DBUser.record_format]"""
        record = {'language': 'fr', 'expiration': None}
        error = DBUser.record_format(record)
        self.assertEqual(len(record), 1)
        self.assertIsInstance(error, TranslationString)
        self.assertIn('without login', error)

        record['login'] = 'test1'
        error = DBUser.record_format(record)
        self.assertIsInstance(error, TranslationString)
        self.assertIn('without password', error)

        record['password'] = 'test1pwd'
        record['password_mustchange'] = True
        error = DBUser.record_format(record)
        self.assertTrue(record['password'].startswith('$'))
        self.assertTrue(record['password_mustchange'])
        self.assertIsInstance(error, TranslationString)
        self.assertIn('without last name', error)

        record['last_name'] = ' AVULEUR '
        record['honorific'] = 'X'
        error = DBUser.record_format(record)
        self.assertEqual(record['last_name'], 'AVULEUR')
        self.assertIsInstance(error, TranslationString)
        self.assertIn('unknown honorific', error)

        record['honorific'] = 'Mrs'
        error = DBUser.record_format(record)
        self.assertIsInstance(error, TranslationString)
        self.assertIn('without email', error)

        record['email'] = 'test1_chrysal.io'
        error = DBUser.record_format(record)
        self.assertIsInstance(error, TranslationString)
        self.assertIn('Invalid email', error)

        record['email'] = 'test1@chrysal.io'
        record['authority_check'] = '2020-01-28T12:00:00'
        error = DBUser.record_format(record)
        self.assertIsNone(error)

    # -------------------------------------------------------------------------
    def test_record_convert_dates(self):
        """[u:models.dbuser.DBUser.record_convert_dates]"""
        record = {
            'last_login': '2020-01-01T12:00:00',
            'password_update': '2020-06-01T12:00:00',
            'expiration': '2020-12-01',
            'account_creation': '2020-01-01T12:00:00',
            'account_update': '2020-06-01T12:00:00',
            'authority_check': '2020-06-01T12:00:00'
        }
        DBUser.record_convert_dates(record)
        self.assertIsInstance(record['last_login'], datetime)
        self.assertIsInstance(record['password_update'], datetime)
        self.assertIsInstance(record['expiration'], date)
        self.assertIsInstance(record['account_creation'], datetime)
        self.assertIsInstance(record['account_update'], datetime)
        self.assertIsInstance(record['authority_check'], datetime)

    # -------------------------------------------------------------------------
    def test_db2xml(self):
        """[u:models.dbuser.DBUser.db2xml]"""
        dbsession = self.request.dbsession
        update_time = (datetime.now() + timedelta(minutes=3))\
            .isoformat().partition('.')[0]
        self.add_user({
            'login': 'test1', 'last_login': '2020-01-01T12:00:00',
            'status': 'inactive', 'password': 'test1pwd',
            'password_mustchange': True, 'first_name': ' Édith',
            'last_name': '  AVULEUR ', 'honorific': 'Mrs',
            'email': 'test1@chrysal.io', 'email_hidden': True,
            'language': 'fr', 'time_zone': 'Europe/Paris',
            'theme': 'Alternative', 'page_size': 80,
            'attachments_key': 'Test1',
            'picture': 'test1.svg', 'expiration': '2020-12-31',
            'account_update': update_time,
            'authority': 'ldap', 'authority_check': datetime.now()})
        dbuser = dbsession.query(DBUser).filter_by(login='test1').first()

        user_elt = dbuser.db2xml(dbsession)
        self.assertEqual(user_elt.get('status'), 'inactive')
        self.assertIsNotNone(user_elt.get('created'))
        self.assertTrue(
            datetime.strptime(user_elt.get('created'), '%Y-%m-%dT%H:%M:%S') >
            datetime.now() - timedelta(seconds=2))
        self.assertIsNotNone(user_elt.get('updated'))
        self.assertTrue(
            datetime.strptime(user_elt.get('updated'), '%Y-%m-%dT%H:%M:%S') >
            datetime.now() + timedelta(minutes=2))

        self.assertEqual(user_elt.findtext('login'), 'test1')

        self.assertIsNotNone(user_elt.find('password'))
        self.assertIsNotNone(user_elt.find('password').get('updated'))
        self.assertTrue(
            datetime.strptime(
                user_elt.find('password').get('updated'),
                '%Y-%m-%dT%H:%M:%S') > datetime.now() - timedelta(seconds=2))
        self.assertTrue(user_elt.find('password').get('must-change'))

        self.assertEqual(user_elt.findtext('firstname'), 'Édith')
        self.assertEqual(user_elt.findtext('lastname'), 'AVULEUR')
        self.assertEqual(user_elt.findtext('honorific'), 'Mrs')

        self.assertIsNotNone(user_elt.find('email'))
        self.assertEqual(user_elt.findtext('email'), 'test1@chrysal.io')
        self.assertTrue(user_elt.find('email').get('hidden'))

        self.assertEqual(user_elt.findtext('language'), 'fr')
        self.assertEqual(user_elt.findtext('timezone'), 'Europe/Paris')
        self.assertEqual(user_elt.findtext('theme'), 'Alternative')
        self.assertEqual(user_elt.findtext('page-size'), '80')
        self.assertEqual(user_elt.findtext('expiration'), '2020-12-31')

    # -------------------------------------------------------------------------
    def test_db2xml_extra(self):
        """[u:models.dbuser.DBUser.db2xml_extra]"""
        from ..models.dbprofile import DBProfile
        from ..models.dbuser import DBUserProfile

        dbsession = self.request.dbsession
        dbsession.add(DBProfile(
            profile_id='user_manager',
            i18n_label=dumps({'en': 'User manager'})))
        self.add_user({
            'login': 'test1', 'first_name': 'Édith', 'last_name': 'AVULEUR',
            'password': 'test1pwd', 'email': 'test1@chrysal.io'})
        dbuser = dbsession.query(DBUser).filter_by(login='test1').first()
        dbuser.profiles.append(DBUserProfile(profile_id='user_manager'))
        user_elt = etree.XML(
            '<user>'
            '  <login>test1</login>'
            '  <password>test1pwd</password>'
            '  <firstname>Édith</firstname>'
            '  <lastname>AVULEUR</lastname>'
            '  <email>test1@chrysal.io</email>'
            '</user>')
        dbuser.db2xml_extra(user_elt)
        self.assertEqual(len(user_elt.findall('profiles/profile')), 1)

    # -------------------------------------------------------------------------
    def test_paging_filter(self):
        """[u:models.dbuser.DBUser.paging_filter]"""
        from ..lib.form import Form
        from ..lib.paging import Paging
        from ..lib.filter import Filter

        dbsession = self.request.dbsession
        paging_id = 'users'
        user_filter = Filter(
            self.request, paging_id, (
                ('login', 'Login', True, None),
                ('status', 'Status', False, None)),
            (('status', '=', 'active'),))
        dbquery = user_filter.sql(dbsession.query(DBUser), 'users')
        user_paging = Paging(self.request, paging_id, dbquery)

        html = DBUser.paging_filter(
            self.request, Form(self.request), user_filter, user_paging)
        self.assertIn('page_size', html)
        self.assertIn('filter_key', html)
        self.assertIn('filter_value', html)

    # -------------------------------------------------------------------------
    def test_tab4view(self):
        """[u:models.dbuser.DBUser.tab4view]"""
        from ..lib.form import Form
        from ..models.dbprofile import DBProfile
        from ..models.dbuser import DBUserProfile

        dbsession = self.request.dbsession
        dbsession.add(DBProfile(
            profile_id='user_creator',
            i18n_label=dumps({'en': 'User Creator'}),
            i18n_description={'en': 'Permission to create user.'}))
        dbsession.add(DBProfile(
            profile_id='user_viewer', i18n_label=dumps({'en': 'User Viewer'})))
        dbuser = self._make_one()
        dbsession.add(dbuser)
        dbsession.flush()
        dbsession.add(DBUserProfile(
            user_id=dbuser.user_id, profile_id='user_creator'))
        dbsession.add(DBUserProfile(
            user_id=dbuser.user_id, profile_id='user_viewer'))
        self.configurator.add_route(
            'profile_view', '/profile/view/{profile_id}')
        self.configurator.add_route('group_view', '/group/view/{group_id}')

        # Information
        html = dbuser.tab4view(self.request, 0, Form(self.request))
        self.assertIn('Édith', html)
        self.assertIn('Mrs', html)

        # Preferences
        html = dbuser.tab4view(self.request, 1, Form(self.request))
        self.assertIn('Lines per page:', html)

        # Profiles
        html = dbuser.tab4view(self.request, 2, Form(self.request))
        self.assertIn('User Creator', html)
        dbuser = self._make_one()
        html = dbuser.tab4view(self.request, 2, Form(self.request))
        self.assertIn('No attributed profile.', html)

        # Groups
        html = dbuser.tab4view(self.request, 3, Form(self.request))
        self.assertIn('user is in any group', html)
        dbuser.groups.append(DBGroup(
            group_id='managers', i18n_label=dumps({'en': 'Managers'})))
        html = dbuser.tab4view(self.request, 3, Form(self.request))
        self.assertIn('Managers', html)

        # Other
        self.assertEqual(
            dbuser.tab4view(self.request, 4, Form(self.request)), '')

    # -------------------------------------------------------------------------
    def test_settings_schema(self):
        """[u:models.dbuser.DBUser.settings_schema]"""
        from ..models.dbuser import DBUserProfile

        profiles = {
            'user_creator': ('User Creator', 'Authorized to create users.')}
        groups = {'managers': ('Managers', 'Thies users manage.')}
        self.request.registry['themes'] = {'Default': {}, 'Alternative': {}}
        self.request.session['user'] = {'user_id': 2}
        self.request.has_permission = \
            lambda x: x in ('user-create', 'user-edit')

        schema, defaults = DBUser.settings_schema(
            self.request, profiles, groups)
        self.assertIsInstance(schema, SchemaNode)
        serialized = schema.serialize()
        self.assertIn('status', serialized)
        self.assertIn('password1', serialized)
        self.assertIn('email', serialized)
        self.assertIn('pfl:user_creator', serialized)
        self.assertIsInstance(defaults, dict)
        self.assertIn('status', defaults)
        self.assertEqual(defaults['status'], 'active')

        dbuser = self._make_one()
        self.request.has_permission = lambda x: True
        self.request.registry['authorities'] = {}
        schema = DBUser.settings_schema(
            self.request, profiles, groups, dbuser)[0]
        self.assertIn('authority', schema.serialize())

        dbuser.profiles.append(DBUserProfile(profile_id='user_creator'))
        dbuser.groups.append(DBGroup(
            group_id='managers', i18n_label=dumps({'en': 'Managers'})))
        self.request.has_permission = lambda x: False
        schema, defaults = DBUser.settings_schema(
            self.request, profiles, groups, dbuser)
        self.assertNotIn('authority', schema.serialize())
        self.assertNotIn('status', defaults)
        self.assertIn('pfl:user_creator', defaults)
        self.assertIn('grp:managers', defaults)

    # -------------------------------------------------------------------------
    def test_tab4edit(self):
        """[u:models.dbuser.DBUser.tab4edit]"""
        from ..lib.form import Form

        self.request.session['user'] = {'user_id': 2}
        self.request.has_permission = \
            lambda x: x in ('user-create', 'user-edit')
        self.request.registry['authorities'] = {'ldap': None}
        form = Form(self.request)

        # Information
        html = DBUser.tab4edit(self.request, 0, form, None, None)
        self.assertIn('Identity', html)
        self.assertIn('Security', html)
        self.assertIn('expiration', html)

        self.request.has_permission = lambda x: True
        html = DBUser.tab4edit(self.request, 0, form, None, None)
        self.assertIn('authority', html)

        self.request.has_permission = lambda x: False
        profiles = {
            'user_creator': ('User Creator', 'Authorized to create users.'),
            'user_editor': ('User Editor', None)}
        groups = {'managers': ('Managers', 'Thies users manage.')}
        dbuser = self._make_one()
        html = DBUser.tab4edit(self.request, 0, form, profiles, groups, dbuser)
        self.assertNotIn('expiration', html)

        dbuser.authority = 'ldap'
        html = DBUser.tab4edit(self.request, 0, form, profiles, groups, dbuser)
        self.assertNotIn('password', html)
        self.assertIn('ldap', html)

        # Preferences
        html = DBUser.tab4edit(self.request, 1, form, {}, {})
        self.assertIn('Hide e-mail:', html)
        self.assertNotIn('theme', html)

        self.request.registry['themes'] = {
            'Default': {'name': {'en': 'Default'}},
            'Alternative': {'name': {'en': 'Alternative Chrysalio'}}}
        self.request.registry['settings'] = {'theme': 'Default'}
        html = DBUser.tab4edit(self.request, 1, form, {}, {})
        self.assertIn('theme', html)

        # Profiles
        html = DBUser.tab4edit(self.request, 2, form, {}, {})
        self.assertIn('No available profile.', html)

        html = DBUser.tab4edit(self.request, 2, form, profiles, {})
        self.assertIn('not have the rigths', html)

        self.request.has_permission = \
            lambda x: x in ('user-create', 'user-edit')
        html = DBUser.tab4edit(self.request, 2, form, profiles, {})
        self.assertIn('pfl:user_creator', html)

        # Groups
        html = DBUser.tab4edit(self.request, 3, form, {}, {})
        self.assertIn('No available group.', html)

        html = DBUser.tab4edit(self.request, 3, form, {}, groups)
        self.assertIn('grp:managers', html)

        self.request.has_permission = lambda x: False
        html = DBUser.tab4edit(self.request, 3, form, {}, groups)
        self.assertIn('not have the rigths', html)

        # Other
        html = DBUser.tab4edit(self.request, 4, form, {}, {})
        self.assertEqual(html, '')


# =============================================================================
class IModelsDBUser(DBUnitTestCase):
    """Integration test class for ``models.dbuser``."""

    # -------------------------------------------------------------------------
    def test_delete_cascade_profile(self):
        """[i:models.dbuser] delete, cascade profile"""
        from ..models.dbprofile import DBProfile
        from ..models.dbuser import DBUserProfile

        dbsession = self.request.dbsession
        dbsession.add(DBProfile(
            profile_id='user_manager',
            i18n_label=dumps({'en': 'User manager'})))
        dbuser = DBUser(
            login='test1', last_name='AVULEUR', email='test1@chrysal.io')
        dbuser.set_password('test1pwd')
        dbsession.add(dbuser)
        dbsession.flush()
        dbsession.add(DBUserProfile(
            user_id=dbuser.user_id, profile_id='user_manager'))

        dbuser = dbsession.query(DBUser).filter_by(login='test1').first()
        self.assertIsInstance(dbuser, DBUser)
        user_id = dbuser.user_id
        self.assertIsNotNone(dbsession.query(DBUserProfile).filter_by(
            user_id=user_id).first())

        dbuser.profiles.remove(dbuser.profiles[0])
        dbsession.flush()
        self.assertIsNone(dbsession.query(DBUserProfile).filter_by(
            user_id=user_id).first())



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``lib.i18n`` functions."""

from unittest import TestCase

from pkg_resources import register_loader_type, DefaultProvider
from _pytest.assertion.rewrite import AssertionRewritingHook
import colander

from pyramid.testing import DummyRequest

# pylint: disable = unused-import
from ..models.dbgroup import DBGroup  # noqa
# pylint: enable = unused-import
from . import DBUnitTestCase


# =============================================================================
class ULibI18nLocaleNegotiator(TestCase):
    """Unit test class for :func:`lib.i18n.locale_negotiator`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.i18n.locale_negotiator]"""
        from webob.acceptparse import AcceptLanguageValidHeader
        from ..lib.i18n import locale_negotiator

        request = DummyRequest(accept_language=AcceptLanguageValidHeader(
            'fr-FR, fr;q=0.8, en-US;q=0.5, en;q=0.3'))
        request.registry.settings = {}
        request.registry['settings'] = {'language': 'en'}
        language = locale_negotiator(request)
        self.assertEqual(language, 'en')

        request.session['lang'] = 'fr'
        language = locale_negotiator(request)
        self.assertEqual(language, 'fr')


# =============================================================================
class ULibI18nAddTranslationDirs(TestCase):
    """Unit test class for :func:`lib.i18n.add_translation_dirs`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from pyramid.testing import setUp
        self.configurator = setUp(settings={})
        register_loader_type(AssertionRewritingHook, DefaultProvider)

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``pyramid.testing.setUp()``."""
        from pyramid.testing import tearDown
        tearDown()

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.i18n.add_translation_dirs]"""
        from ..lib.i18n import add_translation_dirs

        add_translation_dirs(self.configurator, 'chrysalio')
        self.assertRaises(
            SystemExit, add_translation_dirs, self.configurator, 'foo')

        self.configurator.get_settings()['translation_dirs'] = 'bar'
        self.assertRaises(
            SystemExit, add_translation_dirs, self.configurator, 'foo')


# =============================================================================
class ULibI18nTranslate(TestCase):
    """Unit test class for :func:`lib.i18n.translate`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.i18n.translate]"""
        from ..lib.i18n import _, translate

        request = DummyRequest()
        request.session = {'lang': 'fr'}

        self.assertEqual(
            translate(_('Hello ${n},', {'n': 'Mr Foo'}), 'fr'),
            'Bonjour Mr Foo,')
        self.assertEqual(
            translate(_('Hello ${n},', {'n': 'Mr Foo'}), request=request),
            'Bonjour Mr Foo,')
        self.assertIn('Mr Foo', translate(_('Hello ${n},', {'n': 'Mr Foo'})))


# =============================================================================
class ULibI18nTranslateField(TestCase):
    """Unit test class for :func:`lib.i18n.translate_field`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.i18n.translate_field]"""
        from ..lib.i18n import translate_field

        request = DummyRequest()
        request.registry['settings'] = {'language': 'en'}
        self.assertEqual(translate_field(request, None), '')

        i18n_field = {'en': 'User editor', 'fr': 'Éditeur des utilisateurs'}
        self.assertEqual(translate_field(request, i18n_field), 'User editor')

        request.locale_name = 'fr'
        self.assertEqual(
            translate_field(request, i18n_field), 'Éditeur des utilisateurs')

        request.locale_name = 'en'
        request.session['lang'] = 'fr'
        self.assertEqual(
            translate_field(request, i18n_field), 'Éditeur des utilisateurs')

        request.locale_name = 'es'
        request.session['lang'] = 'es'
        self.assertEqual(translate_field(request, i18n_field), 'User editor')


# =============================================================================
class ULibI18nRecordFormatI18n(TestCase):
    """Unit test class for :func:`lib.i18n.record_format_i18n`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.i18n.record_format_i18n]"""
        from json import dumps
        from ..lib.i18n import record_format_i18n

        # Without label
        record = {}
        self.assertFalse(record_format_i18n(record))

        # With label in 2 different ways
        record = {'i18n_label': dumps({'en': 'foo'}), 'label_fr': 'toto'}
        self.assertTrue(record_format_i18n(record))
        self.assertIn('i18n_label', record)
        self.assertIn('en', record['i18n_label'])
        self.assertIn('fr', record['i18n_label'])
        self.assertNotIn('i18n_description', record)

        record['description_en'] = 'Lorem ipsum'
        self.assertTrue(record_format_i18n(record))
        self.assertIn('i18n_description', record)


# =============================================================================
class ULibI18nViewI18nLabels(DBUnitTestCase):
    """Unit test class for :func:`lib.i18n.view_i18n_labels`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.i18n.view_i18n_labels]"""
        from json import dumps
        from ..lib.form import Form
        from ..lib.i18n import view_i18n_labels
        from ..models.dbprofile import DBProfile

        dbprofile = DBProfile(
            profile_id='user_creator',
            i18n_label=dumps({'en': 'Profile manager'}),
            i18n_description={'en': 'Permission to create profiles'})
        self.request.locale_name = 'en'
        html = view_i18n_labels(self.request, Form(self.request), dbprofile)
        self.assertIn('Profile manager', html)
        self.assertIn('Permission to create profiles', html)


# =============================================================================
class ULibI18nSchemaI18nLabels(DBUnitTestCase):
    """Unit test class for :func:`lib.i18n.schema_i18n_labels`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.i18n.schema_i18n_labels]"""
        from ..lib.i18n import schema_i18n_labels
        from ..models import LABEL_LEN, DESCRIPTION_LEN

        schema = colander.SchemaNode(colander.Mapping())
        schema_i18n_labels(self.request, schema, LABEL_LEN, DESCRIPTION_LEN)
        serialized = schema.serialize()
        self.assertIn('label_en', serialized)
        self.assertIn('description_en', serialized)

        self.request.registry['settings']['language'] = 'fr'
        schema = colander.SchemaNode(colander.Mapping())
        schema_i18n_labels(self.request, schema, LABEL_LEN)
        serialized = schema.serialize()
        self.assertIn('label_en', serialized)
        self.assertNotIn('description_en', serialized)


# =============================================================================
class ULibI18nDefaultsI18nLabels(TestCase):
    """Unit test class for :func:`lib.i18n.defaults_i18n_labels`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.i18n.defaults_i18n_labels]"""
        from json import dumps
        from ..lib.i18n import defaults_i18n_labels
        from ..models.dbprofile import DBProfile

        dbprofile = DBProfile(
            profile_id='user_creator',
            i18n_label=dumps({'en': 'Profile manager'}),
            i18n_description={'en': 'Permission to create profiles'})
        defaults = defaults_i18n_labels(dbprofile)
        self.assertIn('label_en', defaults)
        self.assertEqual(defaults['label_en'], 'Profile manager')
        self.assertIn('description_en', defaults)
        self.assertEqual(
            defaults['description_en'], 'Permission to create profiles')


# =============================================================================
class ULibI18nEditI18nLabels(DBUnitTestCase):
    """Unit test class for :func:`lib.i18n.edit_i18n_labels`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.i18n.edit_i18n_labels]"""
        from ..lib.form import Form
        from ..lib.i18n import edit_i18n_labels
        from ..models import LABEL_LEN, DESCRIPTION_LEN

        html = edit_i18n_labels(
            self.request, Form(self.request), LABEL_LEN, DESCRIPTION_LEN)
        self.assertIn('name="label_en"', html)
        self.assertIn('name="description_en"', html)

        html = edit_i18n_labels(
            self.request, Form(self.request), LABEL_LEN, 0)
        self.assertIn('name="label_en"', html)
        self.assertNotIn('name="description_en"', html)

        html = edit_i18n_labels(
            self.request, Form(self.request), 0, DESCRIPTION_LEN)
        self.assertNotIn('name="label_en"', html)
        self.assertIn('name="description_en"', html)



"""Backup database into an XML file."""

from sys import exit as sys_exit
from os import makedirs
from os.path import exists, join, isdir
from shutil import rmtree
from logging import getLogger
from argparse import REMAINDER

from lxml import etree
from transaction import manager as transaction_manager

from ..lib.i18n import _
from ..lib.utils import scandir, copy_content
from ..lib.xml import create_entire_xml
from ..relaxng import RELAXNG
from ..models import get_tm_dbsession
from ..models.dbsettings import DBSettings
from ..models.populate import db2xml
from . import Script


LOG = getLogger(__name__)


# =============================================================================
def main(args=None):
    """Main function."""
    args = Backup.arguments(Backup.argument_parser(), args)
    if args is not None:
        sys_exit(Backup(args, db2xml, RELAXNG).run(args.directory))


# =============================================================================
class Backup(Script):
    """Class to backup database.

    :type  args: argparse.Namespace
    :param args:
        Command line arguments.
    :param function _db2xml:
        Function to save database fields into XML elements. See
        :meth:`~chrysalio.models.populate.db2xml`.
    :param dict relaxng:
        A dictionary describing the main Relax NG with the following keys:
        ``'root'``, ``'version'`` and ``'file'``.
    :param list includes: (optional)
        List of hard coding `includes`.
    :type  dbsession_factory: sqlalchemy.orm.session.sessionmaker
    :param dbsession_factory: (optional)
        Function to create session.
    """

    # -------------------------------------------------------------------------
    def __init__(self, args, _db2xml, relaxng, includes=None,
                 dbsession_factory=None):
        """Constructor method."""
        super(Backup, self).__init__(args, False, includes, dbsession_factory)
        self._db2xml = _db2xml
        self._relaxng = relaxng

    # -------------------------------------------------------------------------
    @classmethod
    def argument_parser(cls, description='Backup database.'):
        """Create an argument parser object to parse command line arguments.

        :param str description: (optional)
            Description of the script saving the database.
        :rtype: argparse.ArgumentParser
        """
        parser = super(Backup, cls).argument_parser(description=description)
        parser.add_argument('directory', help='path to backup directory')
        parser.add_argument(
            '--no-validation', dest='no_validation',
            help='do not validate the result', action='store_true')
        parser.add_argument('extra', nargs=REMAINDER, help='extra options')
        return parser

    # -------------------------------------------------------------------------
    def run(self, directory):
        """Save asked elements.

        :param str directory:
            Path to the backup directory.
        :rtype: int
        :return:
            Exit code.
        """
        # pylint: disable = too-many-return-statements
        if self.registry is None:
            return 1

        with transaction_manager:
            dbsession = get_tm_dbsession(
                self.registry['dbsession_factory'], transaction_manager)
            if not DBSettings.exists(dbsession):
                LOG.warning(self._translate(_('Database is empty!')))
                return 0

            # Browse elements
            elements = self._db2xml(
                dbsession, modules=self.registry['modules'])

            # Prepare output directory
            if not self._prepare_directory(directory):
                return 1

            # Save XML file
            root = create_entire_xml(
                self._relaxng, elements, not self._args.no_validation)
            # pylint: disable = protected-access
            if not isinstance(root, etree._Element):  # pragma: nocover
                LOG.error(root)
                return 1
            # pylint: enable = protected-access
            try:
                filename = join(directory, '{0}.xml'.format(
                    self.registry.settings['site.uid']))
                etree.ElementTree(root).write(
                    filename, pretty_print=True, encoding='utf-8',
                    xml_declaration=True)
            except (OSError, IOError):  # pragma: nocover
                LOG.error(self._translate(
                    _('${f} is write protected.', {'f': filename})))
                return 1

            # Complete operation
            for module_id in self.registry['modules']:
                self.registry['modules'][module_id].activate(
                    self.registry, dbsession)
            for module_id in self.registry['modules']:
                error = self.registry['modules'][module_id].backup(
                    self._args, self.registry, dbsession, directory)
                if error:  # pragma: nocover
                    LOG.error(self._translate(error))
                    return 1

        # Save attachments
        self._copy_attachments(directory)

        return 0

    # -------------------------------------------------------------------------
    def _prepare_directory(self, directory):
        """Create or clean up output directory.

        :param str directory:
            Path to the backup directory.
        :rtype: bool
        """
        if exists(directory) and not isdir(directory):
            LOG.error(self._translate(
                _('${d} must be a directory.', {'d': directory})))
            return False

        if not exists(directory):
            try:
                makedirs(directory)
            except (OSError, IOError):  # pragma: nocover
                LOG.error(self._translate(
                    _('${d} is write protected.', {'d': directory})))
                return False

        for entry in scandir(directory):
            if entry.is_dir():
                rmtree(entry.path)

        return True

    # -------------------------------------------------------------------------
    def _copy_attachments(self, directory):
        """Possibly copy attachment files.

        :param str directory:
            Path to the backup directory.
        """
        attachments = self.registry.settings.get('attachments')
        if attachments and exists(attachments):
            for entry in scandir(attachments):
                if entry.is_dir():
                    copy_content(entry.path, join(directory, entry.name))



# pylint: disable = import-outside-toplevel
"""Tests of ``modes`` functions."""

from . import ConfiguratorTestCase


# =============================================================================
class UModesIncludeme(ConfiguratorTestCase):
    """Unit test class for :func:`modes.includeme`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:modes.includeme]"""
        from ..modes import MODE_HOME, MODE_ADMIN, includeme

        includeme(self.configurator)
        self.assertIn(MODE_HOME, self.configurator.registry['modes'])
        self.assertIn(MODE_ADMIN, self.configurator.registry['modes'])


# =============================================================================
class UModesBeforeRender(ConfiguratorTestCase):
    """Unit test class for :func:`menu.before_render`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:modes.before_render]"""
        from pyramid.testing import DummyRequest
        from ..lib.modes import Modes
        from ..lib.menu import Menu
        from ..modes import MODE_ADMIN, before_render

        self.configurator.add_route('home', '/')
        self.configurator.add_route('backup', '/backup')
        self.configurator.add_route('settings_view', '/settings/view')
        self.configurator.add_route('profile_index', '/profile/index')
        self.configurator.add_route('user_index', '/user/index')
        self.configurator.add_route('group_index', '/group/index')
        self.configurator.add_route('modules_view', '/modules/view')
        request = DummyRequest()
        request.registry['modes'] = [MODE_ADMIN]
        event = {'request': request}
        before_render(event)
        self.assertIn('modes', event)
        self.assertIn('menu', event)
        self.assertIsInstance(event['modes'], Modes)
        self.assertIsInstance(event['menu'], Menu)



# pylint: disable = import-outside-toplevel
"""Tests of ``includes`` classes and functions."""

from os.path import join, dirname

from . import DBUnitTestCase
# pylint: disable = unused-import
from ..includes.modules.models import DBModule  # noqa
# pylint: enable = unused-import


# =============================================================================
class UIncludesLoadIncludes(DBUnitTestCase):
    """Unit test class for :func:`includes.load_includes`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from ..security import PRINCIPALS
        from . import DummyRootFactory

        super(UIncludesLoadIncludes, self).setUp()

        self.configurator.set_root_factory(DummyRootFactory)
        self.configurator.registry['principals'] = list(PRINCIPALS)

    # -------------------------------------------------------------------------
    def test_bad_include(self):
        """[u:includes.load_includes] bad include"""
        from ..includes import load_includes

        settings = self.configurator.get_settings()
        settings['chrysalio.includes'] = 'chrysalio.includes.foo'
        self.assertRaises(
            SystemExit, load_includes, self.configurator)

    # -------------------------------------------------------------------------
    def test_conflict(self):
        """[u:includes.load_includes] conflict"""
        from . import TEST_INI
        from ..includes import load_includes

        settings = self.configurator.get_settings()
        settings.update({
            '__file__': TEST_INI,
            'chrysalio.includes':
            'chrysalio.includes.modules, chrysalio.modules.cioskeleton, '
            'chrysalio.tests.modules.conflict'})
        self.assertRaises(
            SystemExit, load_includes, self.configurator)

    # -------------------------------------------------------------------------
    def test_bad_depencendies(self):
        """[u:includes.load_includes] bad dependencies"""
        from . import TEST_INI
        from ..includes import load_includes

        settings = self.configurator.get_settings()
        settings['__file__'] = TEST_INI
        settings['chrysalio.includes'] = \
            'chrysalio.includes.modules, chrysalio.modules.cioskeleton'
        self.assertRaises(
            SystemExit, load_includes, self.configurator)

    # -------------------------------------------------------------------------
    def test_ok(self):
        """[u:includes.load_includes] ok"""
        from . import TEST_INI
        from ..includes import load_includes
        from ..modules.cioskeleton import ModuleCioSkeleton

        settings = self.configurator.get_settings()
        settings.update({
            'chrysalio.includes':
            'chrysalio.includes.themes, chrysalio.includes.modules, '
            'chrysalio.modules.cioldap, chrysalio.modules.cioskeleton',
            'theme.roots': join(dirname(__file__), '..', 'Themes'),
            'theme.patterns': 'Default'})
        self.configurator.registry['themes'] = {
            'Default': {'name': {'en': 'Default'}}}
        self.configurator.registry['settings']['theme'] = 'Default'
        self.configurator.get_settings()['__file__'] = TEST_INI

        # One module not activated
        load_includes(self.configurator)
        self.assertIn('modules', self.configurator.registry)
        self.assertIn('modules_off', self.configurator.registry)
        self.assertIn(
            'chrysalio.modules.cioskeleton',
            self.configurator.registry['modules'])
        self.assertEqual(len(self.configurator.registry['modules']), 2)
        self.assertIsInstance(
            self.configurator.registry['modules'][
                'chrysalio.modules.cioskeleton'], ModuleCioSkeleton)



"""Bone view callables."""

from pyramid.view import view_config

from ....lib.i18n import _
from ....views import BaseView
from .. import Module


# =============================================================================
class BoneView(BaseView):
    """Class to show how to add views with a module.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """

    # -------------------------------------------------------------------------
    @view_config(
        route_name='bone_index',
        renderer='chrysalio:modules/cioskeleton/Templates/bone_index.pt',
        permission='bone-view')
    def index(self):
        """List all bones."""
        Module.check_activated(self._request, 'chrysalio.modules.cioskeleton')
        self._request.breadcrumbs(_('Bone Management'), 2)
        return {}



"""Settings management view callables."""

from lxml import etree

from pyramid.view import view_config
from pyramid.response import Response
from pyramid.httpexceptions import HTTPFound

from ..lib.i18n import _
from ..lib.log import log_info
from ..lib.utils import make_id
from ..lib.xml import create_entire_xml
from ..lib.form import get_action, Form
from ..lib.tabset import Tabset
from ..models.populate import xml2db, web2db
from ..models.dbsettings import DBSettings
from . import BaseView


# =============================================================================
class SettingsView(BaseView):
    """Class to manage settings views.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """

    # -------------------------------------------------------------------------
    @view_config(
        route_name='settings_view',
        renderer='chrysalio:Templates/settings_view.pt',
        permission='settings-view')
    @view_config(route_name='settings_view', renderer='json', xhr=True)
    def view(self):
        """Show general settings."""
        # Ajax
        i_editor = self._request.has_permission('settings-edit')
        if self._request.is_xhr:
            if i_editor:
                web2db(self._request, xml2db, 'settings')
                self._request.registry['settings'] = DBSettings.db2dict(
                    self._request.registry.settings, self._request.dbsession,
                    self._request.registry['settings']['email'])
                log_info(self._request, 'settings_import')
            return {}

        # Action
        action = get_action(self._request)[0]
        if action == 'exp!':
            action = self._settings2response()
            if action:
                return action
        elif action == 'imp!' and i_editor:
            web2db(self._request, xml2db, 'settings')
            self._request.registry['settings'] = DBSettings.db2dict(
                self._request.registry.settings, self._request.dbsession,
                self._request.registry['settings']['email'])
            log_info(self._request, 'settings_import')

        # Breadcrumbs
        self._request.breadcrumbs(
            _('General Settings'), 2,
            replace=self._request.route_path('settings_edit'))

        return {
            'form': Form(self._request), 'dbsettings': DBSettings,
            'i_editor': i_editor, 'action': action,
            'download_max_size': self._request.registry['settings'][
                'download-max-size'],
            'tabset': Tabset(
                self._request, 'tabSettings',
                DBSettings.settings_tabs(self._request))}

    # -------------------------------------------------------------------------
    @view_config(
        route_name='settings_edit',
        renderer='chrysalio:Templates/settings_edit.pt',
        permission='settings-edit')
    def edit(self):
        """Edit general settings."""
        # Form and action
        form = Form(self._request, *DBSettings.settings_schema(self._request))
        action = get_action(self._request)[0]
        if action == 'sav!' and form.validate():
            self._save(form.values)
            log_info(self._request, 'settings_edit')
            return HTTPFound(self._request.route_path('settings_view'))
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        # Breadcrumbs
        self._request.breadcrumbs(
            _('General Settings Edition'), 2,
            replace=self._request.route_path('settings_view'))

        return {
            'form': form, 'dbsettings': DBSettings,
            'tabset': Tabset(
                self._request, 'tabSettings',
                DBSettings.settings_tabs(self._request))}

    # -------------------------------------------------------------------------
    def _settings2response(self):
        """Export settings as an XML file embedded in a Pyramid response.

        :rtype: :class:`pyramid.response.Response` or ``''``
        """
        # Create the XML file
        root = create_entire_xml(
            self._request.registry['relaxng'],
            [DBSettings.db2xml(self._request.dbsession)])
        # pylint: disable = protected-access
        if not isinstance(root, etree._Element):  # pragma: nocover
            return Response(body=root)
        # pylint: enable = protected-access

        filename = '{0}_settings.{1}.xml'.format(
            make_id(self._request.registry['settings']['title'], 'token'),
            DBSettings.suffix)
        response = Response(
            body=etree.tostring(
                root, pretty_print=True, xml_declaration=True,
                encoding='utf-8'),
            content_type='application/xml')
        response.headerlist.append((
            'Content-Disposition', 'attachment; filename="{0}"'.format(
                filename)))

        log_info(self._request, 'settings_export')
        return response

    # -------------------------------------------------------------------------
    def _save(self, values):
        """Save settings."""
        # Update settings table
        for key, value in values.items():
            dbsetting = self._request.dbsession.query(DBSettings).filter_by(
                key=key).first()
            if dbsetting is None:  # pragma: nocover
                self._request.dbsession.add(DBSettings(key=key, value=value))
            else:
                dbsetting.value = value

        # Update registry
        self._request.registry['settings'] = DBSettings.db2dict(
            self._request.registry.settings, self._request.dbsession,
            values['email'])



"""Base class for modules."""

from os.path import dirname
from configparser import ConfigParser

from lxml import etree

from pyramid.httpexceptions import HTTPForbidden

from ..lib.i18n import _
from ..lib.utils import tounicode


# =============================================================================
class Module(object):
    """Base class for included modules.

    :param str config_ini:
        Absolute path to the configuration file (e.g. development.ini).
    """

    name = None
    implements = ()
    dependencies = ()
    relaxng = None
    xml2db = None
    db2xml = None
    areas = None
    _DBModule = None  # In order to avoid always loading module table

    # -------------------------------------------------------------------------
    def __init__(self, config_ini):
        """Constructor method."""
        # pylint: disable = unused-argument
        self.uid = self.__class__.__module__
        if self.name is None:
            self.name = self.uid
        if self.areas is None:
            self.areas = {self.uid: self.name}

    # -------------------------------------------------------------------------
    @classmethod
    def register(cls, environment, module_class):
        """Method to register the module.

        :type  environment: :class:`pyramid.config.Configurator` or
            :class:`dict`
        :param environment:
            Object used to do configuration declaration within the application
            or a ScriptRegistry to simulate the application registry.
        :param module_class:
            Module class.
        """
        # Server mode (environment == configurator)
        if hasattr(environment, 'registry'):
            if 'modules' in environment.registry:
                module = module_class(environment.get_settings()['__file__'])
                environment.registry['modules'][module.uid] = module

        # Populate/backup/execute mode (environment == ScriptRegistry)
        else:
            module = module_class(environment.settings['__file__'])
            environment['modules'][module.uid] = module

    # -------------------------------------------------------------------------
    @classmethod
    def check_conflicts(cls, includes, modules):
        """Check conflicts between modules. Return a list of IDs of avaliable
        modules and IDs of implemented functionalities.

        :param list includes:
            List of available `includes`.
        :type  modules: collections.OrderedDict
        :param modules:
            Dictionary of available modules.
        :rtype: tuple
        :return:
           A tuple such as ``(implementation_list, error)``.
        """
        implementations = {k: k for k in includes}
        for module_id in modules:
            for implementation in modules[module_id].implements:
                if implementation in implementations:
                    return (), _(
                        '*** Modules ${m1} and ${m2} implement the same '
                        'functionality "${f}".', {
                            'm1': implementations[implementation],
                            'm2': module_id, 'f': implementation})
                implementations[implementation] = module_id
        return implementations.keys(), None

    # -------------------------------------------------------------------------
    def check_dependencies(self, implementations):
        """Check dependencies.

        :param list implementations:
            List containing the IDs of available `includes` plus the IDS
            of functionalities implemented by each available module.
        :rtype: :class:`pyramid.i18n.TranslationString` or ``None``
        """
        for dependency in self.dependencies:
            if dependency not in implementations:
                return _(
                    '*** Dependency "${d}" is missing!', {'d': dependency})
        return None

    # -------------------------------------------------------------------------
    def check_activations(self, modules_off):
        """Check activations.

        :param set modules_off:
            Set of inactive modules.
        :rtype: bool
        :return:
             ``True`` if an activation occurs.
        """
        if self.uid in modules_off:
            return False

        activation = False
        for dependency in self.dependencies:
            if dependency in modules_off:
                modules_off.remove(dependency)
                activation = True
        return activation

    # -------------------------------------------------------------------------
    def module_xml2db(self, dbsession, tree, only, error_if_exists):
        """Load an XML configuration file for the module.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :type  tree: lxml.etree.ElementTree
        :param tree:
            Content of the XML document.
        :param str only:
            If not ``None``, only the items of type ``only`` are loaded.
        :param bool error_if_exists:
            It returns an error if an item already exists.
        :rtype: list
        :return:
            A list of error messages.
        """
        # Module activation
        module_elt = None
        if self._DBModule is not None:
            module_elt = tree.xpath(
                'module[@id="{0}"]|modules/module[@id="{0}"]'.format(
                    self.uid))
            if module_elt:
                dbmodule = dbsession.query(self._DBModule).filter_by(
                    module_id=self.uid).first()
                inactive = module_elt[0].get('inactive') in ('true', '1')
                if inactive and dbmodule is None:
                    # pylint: disable = not-callable
                    dbsession.add(
                        self._DBModule(module_id=self.uid, inactive=True))
                elif not inactive and dbmodule:
                    dbsession.delete(dbmodule)

        # Module settings
        # pylint: disable = unsubscriptable-object
        if self.relaxng is None or self.xml2db is None or (
                module_elt and len(module_elt[0]) + 1 == 1):
            return []
        if module_elt:
            root_elt = module_elt[0].xpath(
                'ns0:{0}'.format(self.relaxng['root']),
                namespaces={'ns0': self.relaxng.get('namespace')})
        else:
            root_elt = tree.xpath(
                '/ns0:{0}'.format(self.relaxng['root']),
                namespaces={'ns0': self.relaxng.get('namespace')})
        if not root_elt:
            return []

        return self.xml2db[0](dbsession, root_elt[0], only, error_if_exists)

    # -------------------------------------------------------------------------
    def module_db2xml(self, dbsession):
        """Return a list of XML elements of the module.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :rtype: list
        :return:
            A list of XML elements.
        """
        module_elt = etree.Element('module')
        module_elt.set('id', self.uid)
        inactive = False

        # Module activation
        if self._DBModule is not None:
            dbmodule = dbsession.query(self._DBModule).filter_by(
                module_id=self.uid).first()
            inactive = dbmodule is not None and dbmodule.inactive
            if inactive:
                module_elt.set('inactive', 'true')

        # Module settings
        # pylint: disable = unsubscriptable-object
        if self.relaxng is None or self.db2xml is None:
            return [module_elt] if inactive else []
        root_elt = etree.SubElement(
            module_elt,
            etree.QName(self.relaxng['namespace'], self.relaxng['root']).text,
            nsmap={None: self.relaxng['namespace']},
            version=self.relaxng['version'])
        self.db2xml[0](dbsession, root_elt)

        return [module_elt]

    # -------------------------------------------------------------------------
    def populate(self, args, registry, dbsession):
        """Method called by populate script to complete the operation.

        :type  args: argparse.Namespace
        :param args:
            Command line arguments.
        :param dict registry:
            Dictionary registry.
        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :rtype: :class:`pyramid.i18n.TranslationString` or ``None``
        :return:
            Error message or ``None``.
        """

    # -------------------------------------------------------------------------
    def backup(self, args, registry, dbsession, directory):
        """Method called by backup script to complete the operation.

        :type  args: argparse.Namespace
        :param args:
            Command line arguments.
        :param dict registry:
            Dictionary registry.
        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :param str directory:
            Path to the backup directory.
        :rtype: :class:`pyramid.i18n.TranslationString` or ``None``
        :return:
            Error message or ``None``.
        """

    # -------------------------------------------------------------------------
    def activate(self, registry, dbsession):
        """Method to activate the module.

        :type  registry: pyramid.registry.Registry
        :param registry:
            Application registry.
        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        """

    # -------------------------------------------------------------------------
    def deactivate(self, registry, dbsession):
        """Method to deactivate the module.

        :type  registry: pyramid.registry.Registry
        :param registry:
            Application registry.
        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        """

    # -------------------------------------------------------------------------
    @classmethod
    def check_activated(cls, request, module_id):
        """Check if the module is active and raise an HTTPForbidden exception
        if not.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str module_id:
            ID of checked module.
        """
        if 'modules' not in request.registry or \
           module_id not in request.registry['modules'] or \
           module_id in request.registry['modules_off']:
            raise HTTPForbidden(comment=_(
                'The module "${i}" is not activated.', {'i': module_id}))

    # -------------------------------------------------------------------------
    def configuration_route(self, request):
        """Return the route to configure this module.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        """

    # -------------------------------------------------------------------------
    def _settings(self, config_ini, section=None):
        """Method to retrieve the settings of the module.

        :param str config_ini:
            Absolute path to the configuration file.
        :param str section: (optional)
            Section in INI file.
        :rtype: dict
        """
        if section is None:
            section = self.__class__.__name__.replace('Module', '')
        config = ConfigParser({'here': dirname(config_ini)})
        config.read(tounicode(config_ini), encoding='utf8')
        if not config.has_section(section):
            return config.defaults()
        return dict(config.items(section))



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``modules.cioskeleton.models.dbbone`` class."""

from ....tests import DBUnitTestCase


# =============================================================================
class UModelsDBBoneDBBone(DBUnitTestCase):
    """Unit test class for
    :class:`modules.cioskeleton.models.dbbone.DBBone`."""

    # -------------------------------------------------------------------------
    def test_xml2db(self):
        """[u:modules.cioskeleton.models.dbbone.DBBone.xml2db]"""
        from lxml import etree
        from pyramid.i18n import TranslationString
        from ..models.dbbone import DBBone

        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member
        dbbone = DBBone(bone_id='bone01', label='Bone n°1')
        dbsession.add(dbbone)

        # error_if_exists
        bone_elt = etree.XML(
            '<bone id="bone01" xmlns="http://ns.chrysal.io/cioskeleton">'
            '  <label>Bone n°1</label>'
            '</bone>')
        error = DBBone.xml2db(dbsession, bone_elt)
        self.assertIsInstance(error, TranslationString)
        error = DBBone.xml2db(
            dbsession, bone_elt, error_if_exists=False)
        self.assertIsNone(error)

        # Bad ID
        bone_elt = etree.XML(
            '<bone id="" xmlns="http://ns.chrysal.io/cioskeleton">'
            '  <label>Bone n°2</label>'
            '</bone>')
        error = DBBone.xml2db(dbsession, bone_elt)
        self.assertIsNotNone(error)

        # Correct bone
        bone_elt = etree.XML(
            '<bone id="bone02" xmlns="http://ns.chrysal.io/cioskeleton">'
            '  <label>Bone n°2</label>'
            '</bone>')
        error = DBBone.xml2db(dbsession, bone_elt)
        self.assertIsNone(error)
        dbbone = dbsession.query(DBBone).filter_by(
            bone_id='bone02').first()
        self.assertIsInstance(dbbone, DBBone)
        self.assertEqual(dbbone.label, 'Bone n°2')

    # -------------------------------------------------------------------------
    def test_db2xml(self):
        """[u:modules.cioskeleton.models.dbbone.DBBone.db2xml]"""
        from ..models.dbbone import DBBone

        dbbone = DBBone(
            bone_id='bone01', label='Bone n°1',
            attachments_key='Bone01', picture='bone01.png')
        bone_elt = dbbone.db2xml()
        self.assertEqual(bone_elt.findtext('label'), 'Bone n°1')
